/* hdrtag.c: this program examines header files and *appends* to tags files
 *  It extracts tags for structure names, typedefs, #defines, and enums.
 *
 *  Author : Dr. Charles E. Campbell, Jr.
 *  Version: 4
 *  History: {{{1
 *           v4     08/29/05 - allows .C suffix to be used (C++)
 *                  01/18/06 - -X flag introduced
 *                  02/27/06 - "automatic" -p for c++ (see main(), fprior(), and hascpparg)
 *           v3     12/11/02 - added bibitem, commands, and environments to tex support
 *                  08/22/03 - fixed bug in string handling for assembly
 *                  10/10/03 - added vim maps (map, vmap, omap, noremap, etc) to vim support
 *                  02/19/04 - fixed \cite for no-commas in \cite{...} list
 *                             Vimmaps pick up lots of non-alphanumeric characters, need to consider
 *                  03/25/04 - turned default warning flag to off
 *                  04/05/04 - -o tagfile and -otagfile supported
 *                  08/03/04 - amended priority tag to label!# instead of label#
 *                  08/20/04 - gives error when applied to unsupported suffix
 *                  12/27/04 - added single-letter tagtypes to Ltagtype for taglist.vim's benefit
 *           v2.28  09/24/02 - changed default of gflag to be true
 *                           - changed default of tflag to be 1
 *           v2.27   3/20/01 - LaTex \cite{..,...,...} now supports commas
 *           v2.26   9/28/01 - vim functions now have duplicates: <SID>func/s:func -> func
 *           v2.25   1/12/01 - #define labels are now Xstate rather than ZeroState (C,C++)
 *           v2.24  10/20/00 - Fixed a bug associated with multi-line strings
 *           v2.23   9/14/00 - Updated internal documentation (hdrtag "?")
 *                               Included vim and maple in initpriority() and setpriority()
 *                               Included -R flag
 *           v2.22f  1/10/00 - fixed etags/TAGS generation
 *           v2.22e 12/22/99 - added command-finding to vimfsa
 *           v2.22d  7/23/99 - removed a few unused variables
 *           v2.22c  7/13/99 - error on Fflag failure included
 *           v2.22b  6/16/99 - began installing proc() and -> Maple tag recognition
 *           v2.22a  4/16/99 - fixed a enum label label (...) problem
 *           v2.22   4/ 7/99 - added function[!] funcname recognition to vimfsa
 *           v2.21   2/ 5/99 - strings now will continue past a line only if \
 *                             terminates that line
 *           v2.20  11/19/98 - included VIM tag generation
 *           v2.19e  4/29/98 - fixed fopenfilename bug (being freed but pointing to stack)
 *           v2.19d  4/16/98 - fixed bug for vms
 *           v2.19c  4/15/98 - fixed bug in strpriority()
 *           v2.19b  3/ 5/98 - include short-funcnames for C++
 *                             improved yacclabel handling
 *           v2.18   2/25/98 - priority -> tagtype, strpriority, -x#
 *                             multiple-xstates removed
 *                             recognizes function names, *except* for functions which
 *                             return (pointers to)\+ function
 *           v2.17   1/30/98 - ignores #error ... lines
 *           v2.16  10/28/97 - now examines #pragma asm ... #pragma endasm for XREF labels
 *           v2.15   9/11/97 - added TeX/LaTeX support
 *           v2.14   6/11/97 - added unix/dodir support
 *                           - broke dodir out and put into xtdio library
 *                           - now permits white space preceding preprocessor statements
 *                           - skips #pragma asm ... #pragma endasm lines
 *           v2.13   6/ 6/97 - added *.l
 *           v2.12  12/23/96 - converted ncurly<0 from error to warning
 *           v2.12  10/23/96 - "-F filename" code installed
 *                   5/ 2/96 - included wflag (warning suppression)
 *                   4/16/96 - included #ifdefctr and #ifdefskip code
 *                  11/ 7/94 - added etags, *.y
 *                   7/13/93 - fixed up class label : label ... { with state 13
 *
 *  What Its Looking For: {{{1
 *
 *    struct LABEL {...};
 *    enum LABEL {
 *      LABEL,...};
 *    #define LABEL ...
 *    typedef ... LABEL,LABEL,...;
 *
 *  And, if gflag is true (default), it looks for C-allowable "words" that
 *    1) are not keywords
 *    2) are outside curly braces and parentheses
 *    3) and, if xflag==1, haven't been seen before
 *
 *  Routines Include: <hdrtag.c> {{{1
 *    char *antimagic(char *line)
 *    char *cname(char *inp,char *cvar)
 *    char *skipcomment(char *b,char *buf,FILE *fp)
 *    char *skipstring(char *b,char *buf,FILE *fp)
 *    int cppaccept(int c)
 *    int main( int argc, char **argv)
 *    int options(char **s,char **a)
 *    int yaccfsa(char *cvar,char *b,char *buf)
 *    int tag_compare(const Label **tag1,const Label **tag2)
 *    int texaccept(int c)
 *    long mergetags(void)
 *    void MaybeFunc(char *b,int *state)
 *    void PrintTag(Label *tag,FILE *fpout)
 *    void dofile(char *filename)
 *    void fprior(int parmqty,PARMDEF *parms)
 *    void freeLabel( Label *label, enum LanguageMode filemode)
 *    void hdretagout(void)
 *    void hdrtag( FILE *fp, char *filename, enum LanguageMode filemode)
 *    void hdrtagout(void)
 *    void initpriority(void)
 *    void makeLabel( char *label, char *line, int fsa_state, unsigned char tagtype)
 *    void asmfsa(char *buf)
 *    void hdrfsa(char *cvar,char *b,char *buf,int ifdefskip)
 *    void setpriority(char *hplabel,int hpval)
 *    void strpriority(char *str)
 *    void texfsa(char *buf)
 *    void vimfsa(char *buf)
 */

/* Header Information: {{{1 */
#include <stdio.h>
#include <ctype.h>
#include "xtdio.h"

/* ------------------------------------------------------------------------
 * Typedefs: {{{2
 */
typedef struct Label_str	Label;

/* ------------------------------------------------------------------------
 * Definitions: {{{2
 */
#define BUFSIZE       6000
#define FFLAG_BUFSIZE (4*BUFSIZE)
#define HPBUFSIZE     128
#define MAXWORD       33
#define TAGBUFSIZE    1024
#define VERSION       "4"

#if defined(MCH_AMIGA) || defined(unix)
# define USE_DODIR
#endif

/* priorities are used by hdrtagout(), qsort(), and tag_compare() to select the
 * highest priority tag amongst several identical tags.  If several tags are
 * identical and have the same priority, which one gets selected is unpredictable.
 *
 * As of 11/7/94:
 *
 *   PTOP   is used only by yaccfsa() so that %tokens will override #defines
 *   PHI    if the parser recognizes a specific type of label, it generally has
 *          PHI priority (typedefs, #defines, structs, etc)
 *   PMED   (see -gflag): for those labels which haven't appeared in the
 *          hdrbase ParSrch table yet and aren't inside ().
 *   PLO    any other accepted label and extern labels
 *   PNEVER is used fprior() to set up keywords such as "extern", "%token"
 *          "class" (for C++ only), etc.  These keywords will never appear
 *          in a <hdrtags> file.
 */
#define PTOP      ((unsigned char)  0)   /* top priority label             */
#define PHIHI     ((unsigned char)  2)   /* high priority label            */
#define PHI       ((unsigned char)  4)   /* high priority label            */
#define PMEDHI    ((unsigned char)  6)   /* medium priority label          */
#define PMED      ((unsigned char)  8)   /* medium priority label          */
#define PMEDLO    ((unsigned char) 10)   /* medium priority label          */
#define PLO       ((unsigned char) 12)   /* low priority label             */
#define PLOLO     ((unsigned char) 14)   /* very low priority label        */
#define PNEVER    ((unsigned char) 16)   /* do not *ever* print this label */
#define PRESTRICT ((unsigned char) 128)  /* this bit indicates restrction  */
#define PRMASK    ((unsigned char) 0x3f) /* this mask removes restriction  */

/* ---------------------------------------------------------------------
 * Enumerations: {{{2
 */
enum ForStates_enum {       /* ForStates: {{{3                                  */
    FtnZeroState=0,         /* 0: zero state                                    */
    ProgramLabel,           /* 1: program                                       */
    FunctionLabel,          /* 2: function                                      */
    SubroutineLabel,        /* 3: subroutine                                    */
    CommonLabel,            /* 4: common block                                  */
    NamelistLabel           /* 5: namelist                                      */
	};
enum HdrStates_enum {       /* HdrStates: {{{3                                  */
    ZeroState=0,            /* 0: zero state                                    */
    ClassLabel,             /* 1: [struct|class] label",                        */
    ClassLabelLabel,        /* 2: [struct|class] label [label|:|{]",         }  */
    EnumLabel,              /* 3: enum label",                                  */
    EnumLabelLabel,         /* 4: enum label label",                            */
    EnumLabelCurly,         /* 5: enum label [{] label [, label...] [}] [;]",   */
    TypedefLabel,           /* 6: typedef label",                               */
    TypedefLabelLabel,      /* 7: typedef label label ...",                     */
    TypedefLabelListStart,  /* 8: typedef label ... [,;{(] ...",             }  */
    TypedefLabelParenStart, /* 9: typedef label ... (...label",                 */
    TypedefLabelParenStop,  /* 10: typedef label ...(... label)",               */
    Extern,                 /* 11: extern",                                     */
    XState,                 /* 12: xstate",                                     */
    ClassLabelColon,        /* 13: class label : ... {",                      } */
    TypedefLabelCurlyStop,  /* 14: typedef label ... { ... }",                  */
    ExternLabelStuff,       /* 15: extern label [;|,|LABEL] ... ;",             */
    TypedefLabelCurly,      /* 16: typedef ... label {,label...};"              */
    TypedefLabelBrace,      /* 17: typedef ... label [,label...];"              */
    EnumLabelResume,        /* 18: enum label ...{...=    (resume on ,)       } */
    MultiZeroState,         /* 19: multi-zero state                             */
    MultiXState             /* 20: multi-xstate                                 */
    };
enum MatStates_enum {		/* MatStates: {{{3                                  */
	matZeroState= 0,        /* 0: zero state                                    */
	matFunction             /* 1: function ...                                  */
	};

enum MplStates_enum {       /* MplStates: {{{3                                  */
	MplZeroState= 0,
	MplMap,
	MplProc,
	MplVar
	};
enum TexStates_enum {       /* TexStates: {{{3                                  */
    texCite= 0,             /* 0: \cite{...}                                    */
    texLabel,               /* 1: \label{...}                                   */
    texBibitem,             /* 2: \bibitem                                      */
    texCommand,             /* 3: \newcommand \renewcommand \providecommand     */
    texEnvironment,         /* 4: \newenvironment \renewenvironment \newtheorem */
    texFunction             /* 5: \Function{...} (cec-ism)                      */
    };
enum VimStates_enum {       /* VimStates: {{{3                                  */
    VimZeroState= 0,        /* 1: zerostate                                     */
    VimSyntax,              /* 2: syn ...                                       */
    VimMatch,               /* 3: syn match ...                                 */
    VimRegion,              /* 4: syn region ...                                */
    VimCluster,             /* 5: syn cluster ...                               */
    VimKeyword,             /* 6: syn keyword ...                               */
    VimFunction,            /* 7: function[!] funcname                          */
    VimFuncName,            /* 8: function name used                            */
    VimCommand,             /* 9: command[!] [-options] command                 */
    VimCmdName,             /* 10: command name used                            */
    VimMap                  /* 11: vim map                                      */
	};
enum YaccStates_enum {      /* YaccStates: {{{3                                 */
    YaccZeroState=0,        /* 0: zerostate                                     */
    YaccKeyWord,            /* 1: yacc-keyword                                  */
    YaccLeft,               /* 2: %%left                                        */
    YaccRight,              /* 3: %%right                                       */
    YaccToken,              /* 4: %%token                                       */
    YaccTokenName,          /* 5: %%token <...>                                 */
    YaccTokenList,          /* 6: %%token <...> ...                             */
    YaccType,               /* 7: %%type                                        */
    YaccTypeName,           /* 8: %%type <...>                                  */
    YaccTypeList,           /* 9: %%type <...> ...                              */
    YaccUnion,              /* 10: %%union                                      */
    YaccPercentCurly,       /* 12: %%{                   }                      */
    YaccLabel,              /* 13: yacclabel                                    */
    YaccStmts,              /* 14: yacclabel : stmnts...                        */
    YaccXState
    };
enum LanguageMode { /* Language: {{{3 */
	BIB,
	C,
	FORTRAN,
	LEX,
	MAPLE,
	MATLAB,
	TEX,
	VIM,
	YACC};
enum TagType {      /* TagTypes: {{{3 */
	TTnever= 0,
	TTclass,
	TTclasslabel,
	TTdefine,
	TTenumlabel,
	TTenumtype,
	TTextern,
	TTglobal,
	TTglobalinparen,
	TTtypedef,
	TTmaybefunc,
	TTshortfuncname,
	TTfuncname,
	TTyacclabel,
	TTyaccleft,
	TTyaccright,
	TTyacctoken,
	TTyacctype,
	TTyaccunionmaybe,
	TTyaccunion,
	TTyaccvar,
	TTtexlabel,
	TTtexcite,
	TTtexfunction,
	TTtexbibitem,
	TTtexcommand,
	TTtexenvironment,
	TTasmxref,
	TTasmxdef,
	TTasmlabel,
	TTasmlabelc,
	TTctags,
	TThdrtags,
	TTvimcmd,
	TTvimfunc,
	TTvimfuncshort,
	TTvimsyntax,
	TTmplproc,
	TTmplmap,
	TTmplvar,
	TTmatfunc,
	TTend			/* this is unused by tagtype, but used to allocate priority */
	};

/* ------------------------------------------------------------------------
 * Local Data Structures: {{{2
 */
struct Label_str {			/* Label:								*/
	enum TagType  tagtype;
	unsigned char fsa_state; /* transit to this state on this label  */
	char         *fname;     /* filename which Label is in           */
	char         *label;     /* Label string                         */
	char         *line;      /* copy of line Label is in             */
	char          isstat;    /* this tag is a "static" tag (-s)      */
	Label        *nxt;       /* linked list support                  */
	Label        *prv;       /* linked list support                  */
	};

/* ------------------------------------------------------------------------
 * Global Data: {{{2
 */
static int   aflag = 1;    /* use antimagic on tags' patterns               */
static int   eflag = 0;    /* generate etags-mode <hdrtags>                 */
static char *Fflag = NULL; /* file contains list of files to be tagged      */
static int   gflag = 1;    /* global variables, prototypes, etc flag        */
static int   hflag = 0;    /* *.h files only                                */
static int   mflag = 0;    /* merge <tags> file with hdrtag output          */
static int   pflag = 0;    /* use C++ keywords, too                         */
static int   Rflag = 0;    /* restrict tags output to given list            */
static int   sflag = 0;    /* for system header files (fixed, static,...)   */
static int   tflag = 1;    /* include tagtype in comment in <hdrtags>       */
static int   vflag = 0;    /* verbose flag                                  */
static int   Vflag = 0;    /* visual mode flag                              */
static int   xflag = 0;    /* exclude duplicates (default is don't exclude) */
static int   Xflag = 0;    /* use exuberant ctags style tagfields           */
static int   wflag = 0;    /* suppress warnings if false                    */
static int   zflag = 0;    /* in <tags> files, precede tags by priority #   */

#ifdef unix
char   *magic      = "^$.\\";
#endif
#ifdef MCH_AMIGA
char   *magic      = "^$.[]|\\";
#endif
#ifdef vms
char   *magic      = "\\";
#endif
#ifdef msdos
char   *magic      = "\\";
#endif


static Label  *maybefunc  = NULL; /* pointer to a ?function? label */
static PS     *forbase    = NULL; /* ParSrch database root         */
static PS     *hdrbase    = NULL; /* for hdrfsa                    */
static PS     *matchbase  = NULL; /* ParSrch database root         */
static PS     *matbase    = NULL; /* for matfsa                    */
static PS     *mplbase    = NULL; /* for mplfsa                    */
static PS     *texbase    = NULL; /* ParSrch database root         */
static PS     *vimbase    = NULL; /* for runvimfsa                 */
static PS     *yaccbase   = NULL; /* for yaccfsa                   */
static char   *fname      = NULL;
static char   *suffices[] = {
	"h"  , "H"   , "hpp", "c"  , "C"    ,
	"cpp", "c++" , "cxx", "ipp", "l"    , "lex"  ,
	"y"  , "yacc", "tex", "sty", "cls"  ,
	"ltx", "bib" , "vim", "mv" , "vimrc",
	"m",
	""};
static long    fsa_lblcnt;				/* count of fsa labels						*/
static int     qCq        = 0;			/* for extern "C" stuff						*/
static int    *priority   = NULL;		/* priority list, will be TTend long		*/
static int     hascpparg  = 0;          /* automatic -p for c++ code support        */

#ifdef USE_DODIR
static int    rflag       = 0;			/* directory recursion flag					*/
#endif


static char   tagfile[TAGBUFSIZE];      /* used to hold <tags> output filename      */
static long   linenum     = 0L;			/* current line counter in file				*/
static long   tagtell     = 0L;			/* current byte position in file			*/
static Label *lhd         = NULL;		/* linked list of Labels (tail pntrs)		*/
static Label *ltl         = NULL;		/* linked list of Labels (tail pntrs)		*/
static char  *Ltagtype[]  = {			/* provides simple TagType -> label			*/
  "TT	-	never",
  "TT	c	class",
  "TT	-	classlabel",
  "TT	d	define",
  "TT	e	enumlabel",
  "TT	e	enumtype",
  "TT	-	extern",
  "TT	v	global",
  "TT	-	globalinparen",
  "TT	t	typedef",
  "TT	-	maybefunc",
  "TT	-	shortfuncname",
  "TT	f	funcname",
  "TT	l	yacclabel",
  "TT	-	yaccleft",
  "TT	-	yaccright",
  "TT	t	yacctoken",
  "TT	T	yacctype",
  "TT	-	yaccunionmaybe",
  "TT	u	yaccunion",
  "TT	-	yaccvar",
  "TT	l	texlabel",
  "TT	-	texcite",
  "TT	f	texfunction",
  "TT	b	texbibitem",
  "TT	c	texcommand",
  "TT	e	texenvironment",
  "TT	-	asmxref",
  "TT	-	asmxdef",
  "TT	-	asmlabel",
  "TT	-	asmlabelc",
  "TT	-	ctags",
  "TT	-	hdrtags",
  "TT	c	vimcmd",
  "TT	f	vimfunc",
  "TT	-	vimfuncshort",
  "TT	s	vimsyntax",
  "TT	-	mplproc",
  "TT	-	mplmap",
  "TT	-	mplvar",
  "TT	-	matfunc",
  "TT	-	end"};

/* ------------------------------------------------------------------------
 * Explanation: {{{2
 */
static char *usage[]= {
	XYELLOW,"hdrtag",XGREEN," [",
#ifdef MCH_AMIGA
	XCYAN,"-aghmMprsvx -Amagic -Ffilename",XGREEN,"]\n",
#else
	XCYAN,"-aghmMpsvx -Amagic -Ffilename",XGREEN,"] [",
	XCYAN,"filenames ",XGREEN,"...]\n",
#endif

	"\n",XWHITE,"Options","\n\n",
	XCYAN,"-a",XGREEN,"   : use antimagic on patterns\n",
	XCYAN,"-Amagic",XGREEN,"   : specify \"magic\" characters for antimagic\n",
	XCYAN,"-e",XGREEN,"   : generate TAGS (use flist -ge)\n",
	XCYAN,"-f tagfile",XGREEN,": same as \"-o tagfile\"\n",
	XCYAN,"-F taginfile",XGREEN,": read file of filenames to be hdrtag'ed\n",
	XCYAN,"-g",XGREEN,"   : don't tag global variables, prototypes, etc\n",
	XCYAN,"-h",XGREEN,"   : restrict hdrtag processing to *.h files\n",
	XCYAN,"-m",XGREEN,"   : merge hdrtags with tags\n",
	XCYAN,"-o tagfile",XGREEN,": specify output <tags> file\n",
	XCYAN,"-p",XGREEN,"   : c++ mode: tag global variables, prototypes, etc\n",
	XCYAN,"-P...",XGREEN,": tagtype=priority or tagtypefile (if multiple, separated by \",\")\n",
	"       Priorities may be: top hihi hi medhi med medlo lo lolo never\n",
	XCYAN,"-t",XGREEN,"   : don't include tagtype comments in <hdrtags>\n",
#ifdef USE_DODIR
	XCYAN,"-r",XGREEN,"   : permit recursive descent down directories\n",
#endif
	XCYAN,"-R...",XGREEN,": restrict tags to given list\n",
	XCYAN,"-s",XGREEN,"   : system header file handler (fixed, static files)\n",
	XCYAN,"-v",XGREEN,"   : verbose mode\n",
	XCYAN,"-w",XGREEN,"   : suppress warnings\n",
	XCYAN,"-x",XGREEN,"   : toggle exclude duplicates (default is to exclude)\n",
	XCYAN,"-z",XGREEN,"   : when creating <tags>, precede tag with priority\n",

	"\n",XWHITE,"Explanation",XGREEN,"\n\n",

	"The ",XYELLOW,"hdrtag",XGREEN,
	" program examines header files and creates a ",XRVGREEN,"tags",XGREEN,"\n",
	"file with information on how to tag to structure names, typedefs,\n",
	"#define labels, enum names, and enumerated types.\n\n",

	"Although hdrtag is primarily intended to handle header files (*.h, *.H),\n",
	"it can also handle C (.c), C++ (.cpp, .c++, .C, .cxx), yacc (.y), lex (.l),\n",
	"vim (.vim), and tex (.tex) files.\n\n",

	"The antimagic function inserts backslashes (",
	XRVGREEN,"\\",XGREEN,") before \"magic\" characters.\n",
	"The default set of magic characters is: ",
	XRVGREEN,"^$.[]|\\",XGREEN," .  If your favorite program\n",
	"needs a different set of magic characters, two methods are provided:\n",
	"use the ",
	XRVGREEN,"-Astring",XGREEN," (or ",
	XRVGREEN,"-A string",XGREEN,") option, where the string is your set of\n",
	"magic characters, or set an environment variable, ",
	XRVGREEN,"TAGMAGIC",XGREEN,", to the magic\n",
	"characters.\n\n",

	XRVGREEN,"HDRTAG",XGREEN,", an environment variable, may be a comma-separated list of\n",
	XRVGREEN,"tagtype=#",XGREEN," or ",XRVGREEN,"tagfile",XGREEN,".  The ",XRVGREEN,"#",XGREEN,
	" is the priority of the associated tagtype.\n",
	"The ",XCYAN,"-P",XGREEN," option may also be used to set tagtype priority.  The tagtypes are:\n\n",

	"asmlabel    define         hdrtags     texenvironment  yacclabel\n",
	"asmlabelc   enumlabel      mplmap      texfunction     yaccleft\n",
	"asmxdef     enumtype       mplproc     texlabel        yaccright\n",
	"asmxref     extern         mplvar      typedef         yacctoken\n",
	"class       funcname       texbibitem  vimcmd          yacctype\n",
	"classlabel  global         texcite     vimfunc         yaccunion\n",
	"ctags       globalinparen  texcommand  vimsyntax       yaccvar\n",

	"\n",XWHITE,"Author",XGREEN,"   : Dr. Charles E. Campbell, Jr.\n",
	     XWHITE,"Version",XGREEN,"  : ",VERSION,
	     "\n",XWHITE,"Copyright",XGREEN,": 1999, 2000, 2001  Charles E. Campbell, Jr.\n",
	"",""};

/* ------------------------------------------------------------------------
 * Prototypes: {{{2
 */
int main( int, char **);                                    /* hdrtag.c */
int options(char **,char **);                               /* hdrtag.c */
void fprior(int,PARMDEF *);                                 /* hdrtag.c */
void dofile(char *);                                        /* hdrtag.c */
static void initpriority(void);                             /* hdrtag.c */
static void strpriority(char *);                            /* hdrtag.c */
static int strpriority2(char *);                            /* hdrtag.c */
static void setpriority(char *,int);                        /* hdrtag.c */
static void hdrtag( FILE *, char *, enum LanguageMode);     /* hdrtag.c */
static char *cname(char *,char *);                          /* hdrtag.c */
static void MaybeFunc(char *,int *);                        /* hdrtag.c */
static void asmfsa(char *);                                 /* hdrtag.c */
static void forfsa(char *);                                 /* hdrtag.c */
static void hdrfsa( char *,  char *,  char *,  int);        /* hdrtag.c */
static char *matprepro(char *,FILE *);                      /* hdrtag.c */
static void matfsa(char *);                                 /* hdrtag.c */
static void mplfsa(char *);                                 /* hdrtag.c */
static void texfsa(char *);                                 /* hdrtag.c */
static void vimfsa(char *);                                 /* hdrtag.c */
static int yaccfsa(char *,char *,char *);                   /* hdrtag.c */
static void makeLabel( char *, char *, int, unsigned char); /* hdrtag.c */
static void freeLabel( Label *, enum LanguageMode);         /* hdrtag.c */
static char *skipstring(char *,char *,FILE *);              /* hdrtag.c */
static char *skipcomment(char *,char *,FILE *);             /* hdrtag.c */
static void hdrtagout(void);                                /* hdrtag.c */
static void PrintTag( Label *, FILE *);                     /* hdrtag.c */
static void hdretagout(void);                               /* hdrtag.c */
static int tag_compare(const Label **,const Label **);      /* hdrtag.c */
static char *antimagic(char *);                             /* hdrtag.c */
static long mergetags(void);                                /* hdrtag.c */
int cppaccept(int);                                         /* hdrtag.c */
int texaccept(int);                                         /* hdrtag.c */

/* ------------------------------------------------------------------------
 * Functions: {{{1
 */

/* main: {{{2 */
int main(
  int    argc,
  char **argv)
{
char *dot      = NULL;
char *envmagic = NULL;
int   iarg;

/* Initialize */
rdcolor();              /* initialize color escape sequences          */
initpriority();         /* set up priority                            */
strcpy(tagfile,"tags"); /* default name of tags output file           */

/* get TAGMAGIC if any */
envmagic= getenv("TAGMAGIC");
if(envmagic && envmagic[0]) magic= envmagic;

/* check for c++ */
hascpparg= 0;
for(iarg= 1; iarg < argc; ++iarg) {
	dot= strrchr(argv[iarg],'.');
	if(dot) {
		if(!strcmp(dot,".c++") || !strcmp(dot,".cpp") || !strcmp(dot,".C") || !strcmp(dot,".hpp") || !strcmp(dot,".ipp")) {
			hascpparg= 1;
			break;
			}
		}
	}

/* call upon the command line parser */
#ifdef USE_DODIR
getarg(argc,argv,0,(PARMDEF *) NULL,usage, options,fprior,dodir,1,"Enter filename(s): ");
#else
getarg(argc,argv,0,(PARMDEF *) NULL,usage, options,fprior,dofile,1,"Enter filename(s): ");
#endif

/* process any files given by a Fflag */
if(Fflag) {
	FILE *Ffp;
	char  buf[FFLAG_BUFSIZE];
	Ffp= fopen(Fflag,"r");
	if(Ffp) {
		char *b;
		while(fgets(buf,FFLAG_BUFSIZE,Ffp)) {
			b= strchr(buf,'#');	/* strip off #...comments */
			if(b) *b= '\0';
			srmtrblk(buf);
			dofile(stpblk(buf));
			}
		fclose(Ffp);
		}
	else error(XTDIO_ERROR,"unable to open file<%s%s%s>\n",YELLOW,sprt(Fflag),GREEN);
	}

/* use labels to write tags */
if(eflag) hdretagout();
else      hdrtagout();


/* cleanup */
if(priority) {
	free((char *) priority);
	}

return 0;
}

/* ------------------------------------------------------------------------ */
/* options: this file processes options  {{{2
 *  Returns: 0 for normal next-character option processing
 *	         + for quantity of subsequent arguments to skip
 *	         - for skip the rest of the current argument
 */
int options(char **s,char **a)
{
int ret=0;


switch(**s) {

case 'a':	/* use antimagic */
	aflag= !aflag;
	break;

case 'A':	/* -Astring or -A string: specify magic characters for antimagic */
	if(!s[0][1]) {
		magic= a[1];
		ret  = 1;
		}
	else {
		magic= *s + 1;
		ret  = -1;
		}
	aflag= 1;
	break;

case 'e':	/* generate etags style */
	eflag= !eflag;
	mflag= sflag= 0;
	xflag= 1;
	break;

case 'F':	/* -F filename: read a file of filenames to be hdrtag'ged */
	if(!s[0][1]) {
		Fflag= a[1];
		ret  = 1;
		}
	else {
		Fflag= *s + 1;
		ret  = -1;
		}
	ga_dofile= 0;
	break;

case 'g':	/* global variables, prototypes, etc flag */
	gflag= !gflag;
	if(pflag) gflag= 1;
	break;

case 'h':	/* restrict hdrtag processing to *.h files only */
	hflag= !hflag;
	break;

case 'm':	/* merge with pre-existing tags file */
	mflag= !mflag;
	break;

case 'f':	/* -f tagsfile : give an alternate name for a tags output file */
case 'o':	/* -o tagsfile : give an alternate name for a tags output file */
	if(s[0][1]) { /* -otagfile */
		strcpy(tagfile,s[0]+1);
		ret  = -1;
		}
	else {			/* -o tagfile */
		strcpy(tagfile,a[1]);
		ret  = 1;
		}
	if(!tagfile[0]) strcpy(tagfile,"tags");
	break;

case 'p':	/* use C++ keywords (if editing this, also fix fprior()'s "automatic" -p for c++) */
	pflag= !pflag;
	if(pflag) gflag= 1;
	break;

case 'P':	/* set priority strings: -Pstr=#,str=#,... -Pfilename */
	strpriority(*s + 1);
	ret= -1;
	break;

#ifdef USE_DODIR
case 'r':	/* recursion flag */
	rflag= !rflag;
	break;
#endif

case 'R':	/* restrict tags to given list */
	Rflag= 1;
	if(Rflag && priority) {
		int i;
		for(i= 1; i < TTend; ++i) priority[i]|= PRESTRICT;	/* skip priority[TTnever] -- its always never! */
		strpriority(*s+1);
		}
	ret= -1;
	break;

case 's':	/* system header files flag (uses ###G instead of /...) */
	sflag= !sflag;
	break;

case 't':	/* -t, -tt: include tagtype [and priority] in comment in <hdrtags> */
	if(++tflag >= 3) tflag= 0;
	break;

case 'v':	/* verbose flag */
	if(++vflag > 2) vflag= 0;
	break;

case 'V':	/* visual -- what lines are run through the FSA, what not? */
	Vflag= !Vflag;
	break;

case 'w':	/* suppress warnings if false */
	wflag= !wflag;
	break;

case 'x':	/* exclude duplicates flag */
	/* note: *duplicates* with priority > xflag will be rejected, except for
	 * xflag==0: in that case, include all
	 */
	if(isdigit(s[0][1])) {
		sscanf(*s + 1,"%d",&xflag); /* accept priority <= xflag        */
		ret= -1;                    /* skip rest of arg                */
		}
	else if(xflag) xflag= 0;        /* set exclusion off (include all) */
	else           xflag= PMED;     /* accept priority <= PMED         */
	break;

case 'X':	/* use exuberant-ctags style tagfields */
	Xflag= !Xflag;
	break;

case 'z':	/* precede tags in <tags> files with priority number for sorting purposes */
	/* the idea is to allow sorting later, but at the end the priorities will be stripped */
	zflag= !zflag;
	break;


default:
	if(wflag) error(XTDIO_WARNING,"option <%s%s%s> not currently supported\n", RVGREEN,cprt(**s),GREEN);
	break;
	}

return ret;
}

/* ------------------------------------------------------------------------ */
/* fprior: this function processes prior to dofile {{{2 */
void fprior(int parmqty,PARMDEF *parms)
{
static char *ckeys[]={
	"auto",
	"break",
	"case",
	"char",
	"const",
	"continue",
	"default",
	"do",
	"double",
	"else",
	"entry",
	"float",
	"for",
	"goto",
	"if",
	"int",
	"long",
	"register",
	"return",
	"short",
	"sizeof",
	"static",
	"switch",
	"unsigned",
	"void",
	"while",
	"FILE",
	"NULL",
	""};
static char *cppkeys[]={
	"catch",
	"delete",
	"friend",
	"new",
	"private",
	"protect",
	"public",
	"template",
	"this",
	"throw",
	""};
static char  *vimsyn[] = {"sy","syn","synt","synta","syntax",""};
static char  *vimfun[] = {"fu","fun","func","funct","functi","functio","function",""};
static char  *vimcmd[] = {"com","comm","comma","comman","command",""};
#ifdef USE_VIMMAP
static char  *vimmap[] = {"map",
						  "nm","nma","nmap",
						  "vm","vma","vmap",
						  "om","oma","omap",
						  "im","ima","imap",
						  "lm","lma","lmap",
						  "cm","cma","cmap",
						  "no","nor","nore","norem","norema","noremap",
						  "nno","nnor","nnore","nnorem","nnorema","nnoremap",
						  "vno","vnor","vnore","vnorem","vnorema","vnoremap",
						  "ono","onor","onore","onorem","onorema","onoremap",
						  "ino","inor","inore","inorem","inorema","inoremap",
						  "lno","lnor","lnore","lnorem","lnorema","lnoremap",
						  "cno","cnor","cnore","cnorem","cnorema","cnoremap",
						  ""};
#endif	/* USE_VIMMAP */
char        **ckey     = NULL;
Label        *lbl      = NULL;


/* if hascpparg, but pflag is zero, set pflag */
pflag= gflag= 1;

/* C: enter most C and C++ keywords into hdr ParSrch table */
matchbase= hdrbase= ps_make();
if(pflag) hdrbase->accept= cppaccept;
for(ckey= ckeys; **ckey; ++ckey) makeLabel(*ckey,NULL,(int) ZeroState,TTnever);
if(pflag) for(ckey= cppkeys; **ckey; ++ckey) makeLabel(*ckey,NULL,(int) ZeroState,TTnever);

/* C: enter "special" Finite State Automaton C/C++ keywords */
if(pflag) makeLabel("class", NULL, (int) ClassLabel  , TTnever);
makeLabel("enum"   , NULL, (int) EnumLabel   , TTnever);
makeLabel("struct" , NULL, (int) ClassLabel  , TTnever);
makeLabel("typedef", NULL, (int) TypedefLabel, TTnever);
makeLabel("union"  , NULL, (int) ClassLabel  , TTnever);
makeLabel("extern" , NULL, (int) Extern      , TTnever);

/* FOR: enter most fortran keywords into forbase ParSrch table */
matchbase= forbase= ps_make();
makeLabel("program"   , NULL, (int) ProgramLabel   , TTnever);
makeLabel("function"  , NULL, (int) FunctionLabel  , TTnever);
makeLabel("subroutine", NULL, (int) SubroutineLabel, TTnever);
makeLabel("common"    , NULL, (int) CommonLabel    , TTnever);
makeLabel("namelist"  , NULL, (int) NamelistLabel  , TTnever);

/* MAPLE: enter Finite State Automaton Maple keywords */
matchbase= mplbase= ps_make();
makeLabel("proc" , NULL, (int) MplProc , TTnever);

/* MATLAB: enter Finite State Automaton Matlab keywords */
matchbase= matbase= ps_make();
makeLabel("function" , NULL, (int) matFunction , TTnever);

/* TEX: enter TeX keywords into texbase (ParSrch table) */
matchbase       = texbase    = ps_make();
texbase->accept = texaccept;
makeLabel("cite"            , NULL, (int) texCite       , TTnever);
makeLabel("Function"        , NULL, (int) texCite       , TTnever);
makeLabel("label"           , NULL, (int) texLabel      , TTnever);
makeLabel("bibitem"         , NULL, (int) texBibitem    , TTnever);
makeLabel("newcommand"      , NULL, (int) texCommand    , TTnever);
makeLabel("renewcommand"    , NULL, (int) texCommand    , TTnever);
makeLabel("providecommand"  , NULL, (int) texCommand    , TTnever);
makeLabel("newenvironment"  , NULL, (int) texEnvironment, TTnever);
makeLabel("renewenvironment", NULL, (int) texEnvironment, TTnever);
makeLabel("newtheorem"      , NULL, (int) texEnvironment, TTnever);

/* VIM: enter Finite State Automaton Vim keywords */
matchbase= vimbase= ps_make();
for(ckey= vimsyn; **ckey; ++ckey) makeLabel(*ckey, NULL, (int) VimSyntax  , TTnever);
for(ckey= vimfun; **ckey; ++ckey) makeLabel(*ckey, NULL, (int) VimFunction, TTnever);
for(ckey= vimcmd; **ckey; ++ckey) makeLabel(*ckey, NULL, (int) VimCommand , TTnever);
#ifdef USE_VIMMAP
for(ckey= vimmap; **ckey; ++ckey) makeLabel(*ckey,NULL,(int) VimMap,TTnever);
#endif
makeLabel("cluster", NULL, (int) VimCluster, TTnever);
makeLabel("keyword", NULL, (int) VimKeyword, TTnever);
makeLabel("match"  , NULL, (int) VimMatch  , TTnever);
makeLabel("region" , NULL, (int) VimRegion , TTnever);

/* YACC: enter Yacc keywords into yacc ParSrch table */
matchbase= yaccbase= ps_make();
makeLabel("left" , NULL, YaccLeft , TTnever);
makeLabel("right", NULL, YaccRight, TTnever);
makeLabel("token", NULL, YaccToken, TTnever);
makeLabel("type" , NULL, YaccType , TTnever);
makeLabel("union", NULL, YaccUnion, TTnever);

/* record qty of fsa labels */
for(lbl= lhd, fsa_lblcnt= 0; lbl; lbl= lbl->nxt) ++fsa_lblcnt;

/* optimize the ParSrch trees */
ps_optimize(forbase);
ps_optimize(hdrbase);
ps_optimize(mplbase);
ps_optimize(texbase);
ps_optimize(vimbase);
ps_optimize(yaccbase);

#ifdef USE_DODIR
/* set up dodir */
setup_dodir(dofile,suffices,rflag);
#endif

}

/* ------------------------------------------------------------------------ */
/* dofile: this function processes the command line filenames {{{2 */
void dofile(char *filename)
{
FILE              *fp         = NULL;
enum LanguageMode  filemode   = C;


#ifdef USE_DODIR
/* null filename means dodir has finished processing a directory */
if(!filename) {
	ps_optimize(matchbase);
	return;
	}
#endif

if(!strcmp(filename,"stdin")) {
	fp= stdin;
	stralloc(fname,filename,"dofile: fname allocation problem\n");
	}

else {
	char **suffix;
	char  *dot;

	/* attempt to open filename as-is */
#ifdef __WIN32__
	fp= cygfopen(filename,"r");
#else
	fp= fopen(filename,"r");
#endif
	if(!fp) { /* attempt to open up the filename with a variety of suffices */
		for(suffix= suffices; **suffix; ++suffix) {
			fp= fopenx(filename,"r",*suffix,"");

			if(fp) { /* squirrel away actual filename into fname */
				stralloc(fname,fopenfilename,"dofile: fname allocation problem\n");
				break;
				}
			}
		}
	else {
		stralloc(fname,filename,"failure while allocating fopenfilename\n");
		}

	if(!fp) {		/* unable to open the file ;( */
		fname= NULL;
		if(wflag) error(XTDIO_WARNING,"unable to open <%s%s%s> for reading\n",RVGREEN,filename,GREEN);
		return;
		}

	/* check for lex/yacc/Maple V/vim/tex etc. (ie. an acceptable suffix) */
	dot= strrchr(fname,'.');
	if(dot) {
		if     (!strcmp(dot,".c")   || !strcmp(dot,".h"))    filemode = C;
		else if(!strcmp(dot,".C"))                           filemode = C;
		else if(!strcmp(dot,".cpp") || !strcmp(dot,".hpp"))  filemode = C;
		else if(!strcmp(dot,".ipp"))                         filemode = C;
        else if(!strcmp(dot,".y")   || !strcmp(dot,".yacc")) filemode = YACC;    /* check for yacc file    */
        else if(!strcmp(dot,".l")   || !strcmp(dot,".lex"))  filemode = LEX;     /* check for lex file     */
        else if(!strcmp(dot,".tex") || !strcmp(dot,".sty"))  filemode = TEX;     /* check for tex file     */
        else if(!strcmp(dot,".cls") || !strcmp(dot,".ltx"))  filemode = TEX;     /* check for tex file     */
        else if(!strcmp(dot,".bib"))                         filemode = BIB;     /* check for bibtex file  */
        else if(!strcmp(dot,".vim"))                         filemode = VIM;     /* check for vim file     */
        else if(!strcmp(dot,".vimrc"))                       filemode = VIM;     /* check for vim file     */
        else if(!strcmp(dot,".m"))                           filemode = MATLAB;  /* check for matlab file  */
        else if(!strcmp(dot,".mv"))                          filemode = MAPLE;   /* check for maple file   */
        else if(!strcmp(dot,".for"))                         filemode = FORTRAN; /* check for fortran file */
        else if(!strcmp(dot,".f77"))                         filemode = FORTRAN; /* check for fortran file */
        else if(!strcmp(dot,".ftn"))                         filemode = FORTRAN; /* check for fortran file */
		else error(XTDIO_ERROR,"suffix<%s> not supported\n",sprt(dot));
		}
	}

/* process the file */
hdrtag(fp,fname,filemode);

/* close the file */
if(fp != stdin) fclose(fp);

}

/* --------------------------------------------------------------------- */
/* initpriority: this function initializes the priority table {{{2 */
static void initpriority(void)
{
char *hdrtagpriority= NULL;


priority= (int *) calloc((size_t) TTend,sizeof(int));
if(!priority) error(XTDIO_ERROR,"unable to allocate %d ints for priority!!!\n",TTend);

/* initialize default priorities */
priority[TTnever]         = PNEVER;
priority[TTclass]         = PHI;
priority[TTclasslabel]    = PMEDHI;
priority[TTdefine]        = PHI;
priority[TTenumlabel]     = PMED;
priority[TTenumtype]      = PMED;
priority[TTextern]        = PLO;
priority[TTglobal]        = PMEDLO;
priority[TTglobalinparen] = PLO;
priority[TTmaybefunc]     = PMED;
priority[TTshortfuncname] = PMED;
priority[TTfuncname]      = PHI;
priority[TTtypedef]       = PMEDHI;
priority[TTyacclabel]     = PMEDHI;
priority[TTyaccleft]      = PMED;
priority[TTyaccright]     = PMED;
priority[TTyacctoken]     = PMED;
priority[TTyacctype]      = PMED;
priority[TTyaccunionmaybe]= PNEVER;
priority[TTyaccunion]     = PMED;
priority[TTyaccvar]       = PMED;
priority[TTtexlabel]      = PMED;
priority[TTtexcite]       = PMED;
priority[TTtexfunction]   = PMED;
priority[TTtexbibitem]    = PMED;
priority[TTtexcommand]    = PMED;
priority[TTtexenvironment]= PMED;
priority[TTasmxref]       = PLO;
priority[TTasmxdef]       = PMEDLO;
priority[TTasmlabel]      = PLO;
priority[TTasmlabelc]     = PMED;
priority[TTctags]         = PMED;
priority[TThdrtags]       = PMED;
priority[TTvimfunc]       = PMED;
priority[TTvimfuncshort]  = PMEDLO;
priority[TTvimsyntax]     = PMED;

/* translate HDRTAG TTtag=#,... */
#ifdef vms
hdrtagpriority= getenv("HDRTAGDAT");
#else
hdrtagpriority= getenv("HDRTAG");
#endif
if(hdrtagpriority) strpriority(hdrtagpriority);

}

/* --------------------------------------------------------------------- */
/* strpriority: this function sets priority using a hdrtag priority string {{{2 */
static void strpriority(char *str)
{
char *hp;
char *equal;


/* sanity check */
if(!str) {
	return;
	}

for(hp= strtok(str,","); hp; hp= strtok((char *) NULL,",")) {
	equal= strchr(hp,'=');

	/* handle -P[...,]label=priority[,...] */
	if(equal) setpriority(hp,strpriority2(stpblk(equal+1)));

	else if(!Rflag) { /* handle -P[...,]filename[,...] */
		FILE *hpfp= NULL;
		static char  hpbuf[HPBUFSIZE];
		hpfp= fopen(hp,"r");
		if(hpfp) {
			while(ncgets(hpbuf,HPBUFSIZE,hpfp)) {
				equal= strchr(hp,'=');
				if(!*equal) continue;
				srmtrblk(hpbuf);
				setpriority(hpbuf,strpriority2(stpblk(equal+1)));
				}
			fclose(hpfp);
			}
		}

	/* sets all restrictions to medium priority unless there's an explicity setting */
	else setpriority(hp,PMED);
	}

}

/* --------------------------------------------------------------------- */
/* strpriority2: this function converts {{{2
 *   ####                  to an integer
 *   priority-level-string to an integer
 */
static int strpriority2(char *spriority)
{
int hpval= PNEVER;


if     (isdigit(*spriority)) sscanf(spriority,"%d",&hpval);
else if(!strcmp(spriority,"TOP")   || !strcmp(spriority,"top"))   hpval= PTOP;
else if(!strcmp(spriority,"HIHI")  || !strcmp(spriority,"hihi"))  hpval= PHIHI;
else if(!strcmp(spriority,"HI")    || !strcmp(spriority,"hi"))    hpval= PHI;
else if(!strcmp(spriority,"MEDHI") || !strcmp(spriority,"medhi")) hpval= PMEDHI;
else if(!strcmp(spriority,"MED")   || !strcmp(spriority,"med"))   hpval= PMED;
else if(!strcmp(spriority,"MEDLO") || !strcmp(spriority,"medlo")) hpval= PMEDLO;
else if(!strcmp(spriority,"LO")    || !strcmp(spriority,"lo"))    hpval= PLO;
else if(!strcmp(spriority,"LOLO")  || !strcmp(spriority,"lolo"))  hpval= PLOLO;
else if(!strcmp(spriority,"NEVER") || !strcmp(spriority,"never")) hpval= PNEVER;
else {
	error(XTDIO_WARNING,"strpriority2: I don't understand <%s>\n",sprt(spriority));
	hpval= -1;
	}

return hpval;
}

/* --------------------------------------------------------------------- */
/* setpriority: this function supports letting the user set priority for {{{2
 *              the various tagtypes
 */
static void setpriority(char *hplabel,int hpval)
{

/* sanity check */
if(hpval < 0) {
	return;
	}

if     (!strcmp(hplabel,"class"))         priority[TTclass]         = hpval;
else if(!strcmp(hplabel,"classlabel"))    priority[TTclasslabel]    = hpval;
else if(!strcmp(hplabel,"define"))        priority[TTdefine]        = hpval;
else if(!strcmp(hplabel,"enumlabel"))     priority[TTenumlabel]     = hpval;
else if(!strcmp(hplabel,"enumtype"))      priority[TTenumtype]      = hpval;
else if(!strcmp(hplabel,"extern"))        priority[TTextern]        = hpval;
else if(!strcmp(hplabel,"global"))        priority[TTglobal]        = hpval;
else if(!strcmp(hplabel,"globalinparen")) priority[TTglobalinparen] = hpval;
else if(!strcmp(hplabel,"maybefunc"))     priority[TTmaybefunc]     = hpval;
else if(!strcmp(hplabel,"shortfuncname")) priority[TTshortfuncname] = hpval;
else if(!strcmp(hplabel,"funcname"))      priority[TTfuncname]      = hpval;
else if(!strcmp(hplabel,"typedef"))       priority[TTtypedef]       = hpval;
else if(!strcmp(hplabel,"yacclabel"))     priority[TTyacclabel]     = hpval;
else if(!strcmp(hplabel,"yaccleft"))      priority[TTyaccleft]      = hpval;
else if(!strcmp(hplabel,"yaccright"))     priority[TTyaccright]     = hpval;
else if(!strcmp(hplabel,"yacctoken"))     priority[TTyacctoken]     = hpval;
else if(!strcmp(hplabel,"yacctype"))      priority[TTyacctype]      = hpval;
else if(!strcmp(hplabel,"yaccunion"))     priority[TTyaccunion]     = hpval;
else if(!strcmp(hplabel,"yaccvar"))       priority[TTyaccvar]       = hpval;
else if(!strcmp(hplabel,"texlabel"))      priority[TTtexlabel]      = hpval;
else if(!strcmp(hplabel,"texcite"))       priority[TTtexcite]       = hpval;
else if(!strcmp(hplabel,"texfunction"))   priority[TTtexcite]       = hpval;
else if(!strcmp(hplabel,"texbibitem"))    priority[TTtexcite]       = hpval;
else if(!strcmp(hplabel,"texcommand"))    priority[TTtexcite]       = hpval;
else if(!strcmp(hplabel,"texenvironemtn"))priority[TTtexcite]       = hpval;
else if(!strcmp(hplabel,"asmxref"))       priority[TTasmxref]       = hpval;
else if(!strcmp(hplabel,"asmxdef"))       priority[TTasmxdef]       = hpval;
else if(!strcmp(hplabel,"asmlabel"))      priority[TTasmlabel]      = hpval;
else if(!strcmp(hplabel,"asmlabelc"))     priority[TTasmlabelc]     = hpval;
else if(!strcmp(hplabel,"ctags"))         priority[TTctags]         = hpval;
else if(!strcmp(hplabel,"hdrtags"))       priority[TThdrtags]       = hpval;
else if(!strcmp(hplabel,"vimcmd"))        priority[TTvimcmd]        = hpval;
else if(!strcmp(hplabel,"vimfunc"))       priority[TTvimfunc]       = hpval;
else if(!strcmp(hplabel,"vimfuncshort"))  priority[TTvimfuncshort]  = hpval;
else if(!strcmp(hplabel,"vimsyntax"))     priority[TTvimsyntax]     = hpval;
else if(!strcmp(hplabel,"mplproc"))       priority[TTmplproc]       = hpval;
else if(!strcmp(hplabel,"mplmap"))        priority[TTmplmap]        = hpval;
else if(!strcmp(hplabel,"mplvar"))        priority[TTmplvar]        = hpval;
else if(!strcmp(hplabel,"matfunc"))       priority[TTmatfunc]       = hpval;
else error(XTDIO_WARNING,"cannot set priority with <%s=%d>\n",hplabel,hpval);

}

/* --------------------------------------------------------------------- */
/* hdrtag: extracts tags for structure names, typedefs, #defines, and enums {{{2 */
static void hdrtag(
  FILE             *fp,
  char             *filename,
  enum LanguageMode filemode)
{
char *b             = NULL;
char  buf[BUFSIZE];
char  cvar[BUFSIZE];
int   do_asm        = 0;
int   ifdefctr      = 0;
int   ifdefskip     = 0;
int   lexskip       = 1;
int   lexpp         = 0;
int   ret;


if(hflag) {	/* do not hdrtag files which do not end in .h */
	int flen;
	flen= strlen(filename);
	if(flen <= 2               || filename[flen-2] != '.' ||
	   filename[flen-1] != 'h' || filename[flen-1] != 'H') {
		if(vflag) fprintf(stderr,"%shdrtag %s%s (skipped)\n",CYAN,GREEN,filename);
		return;
		}
	}

/* verbose mode */
if(vflag) fprintf(stderr,"%shdrtag %s%s\n",CYAN,GREEN,filename);

/* initialize on a per-file basis */
linenum  = 0L;										/* reset line counter	*/
tagtell  = 0L;
matchbase= (filemode == C || filemode == LEX)? hdrbase  :
		   (filemode == FORTRAN)?              forbase  :
		   (filemode == MAPLE)?                mplbase  :
		   (filemode == MATLAB)?               matbase  :
           (filemode == TEX)?                  texbase  :
		   (filemode == VIM)?                  vimbase  :
           (filemode == YACC)?                 yaccbase : NULL;
if(!matchbase) error(XTDIO_ERROR,"unable to choose correct language base\n");

while(tagtell= ftell(fp), fgets(buf,BUFSIZE,fp)) {
	++linenum;

	/* lex file processing -- basically, runs it through the usual hdrtag if lexskip is false */
	switch(filemode) {

	case FORTRAN:	/* fortran file processing */
		forfsa(buf);
		break;

	case LEX:	/* lex file processing */
		if(buf[0] == '%' && buf[1] == '{') {
			lexskip= 0;
			if(buf[2] == '\n') continue;
			memcpy(buf,buf+2,strlen(buf+2) + 1);
			}
		else if(buf[0] == '%' && buf[1] == '}') lexskip= 1;
		else if(buf[0] == '%' && buf[1] == '%') {
			if(++lexpp == 2) lexskip= 0;
			continue;
			}
		if(lexskip) continue;
		break;

	case MAPLE:	/* vim file processing */
		mplfsa(buf);
		continue;

	case MATLAB: /* matlab/freespace file processing */
		matfsa(matprepro(buf,fp));
		continue;

	case TEX: /* tex file processing */
		texfsa(buf);
		continue;

	case VIM:	/* vim file processing */
		vimfsa(buf);
		continue;

	default:
		break;
		}

	/* #define processing: ^[ \t]*#... */
	if(*(stpblk(buf)) == '#') {

		if(Vflag) printf("%s%4ld: (%2d,%2d) %s%s",CYAN,linenum,ifdefctr,ifdefskip,YELLOW,buf);
		b= cname(stpblk(buf)+1,cvar);
		if(!strcmp(cvar,"define")) {	/* make an XXX Label: #define XXX */
			b= cname(b+1,cvar);
			makeLabel(cvar,buf,(int) XState,TTdefine);
			}

		if(!strcmp(cvar,"ifdef") || !strcmp(cvar,"ifndef") || !strcmp(cvar,"if")) {
			++ifdefctr;
			}
		else if(!strcmp(cvar,"else") || !strcmp(cvar,"elif")) {
			if(ifdefskip == 0) ifdefskip= ifdefctr;
			continue;
			}
		else if(!strcmp(cvar,"pragma")) {
			b= cname(stpblk(b+1),cvar);	/* cname left b pointer to end-of-cvar character */

			/* skip this line and lines following until #pragma endasm */
			if     (!strcmp(cvar,"asm"))    do_asm= 1;
			else if(!strcmp(cvar,"endasm")) do_asm= 0;
			}
		else if(!strcmp(cvar,"endif")) {
			if(--ifdefctr < 0) {
				if(wflag) error(XTDIO_WARNING,"hdrtag <%s:%ld>: too many #endif's\n",sprt(fname),linenum);
				ifdefctr = 0;
				ifdefskip= 0;
				}
			else if(ifdefctr < ifdefskip) ifdefskip= 0;
			}
		else if(!strcmp(cvar,"error")) {	/* handle #error ... */
			continue;
			}

		while(1) {	/* handle preprocessor continuation lines */
			for(++b; *b; b= stpblk(b+1)) {

				/* handle strings */
				if(!do_asm && (*b == '"' || *b == '\'')) {
					b= skipstring(b,buf,fp);
					if(!*b) break;
					}

				/* handle comments */
				if(do_asm && b[0] == ';') break;
				else if(b[0] == '/') {
					if(b[1] == '/') {
						break; /* C++ style comment, skip line */
						}
					if(b[1] == '*') b= skipcomment(b+1,buf,fp);
					}
				}
			if(do_asm) break;

			if(*b) b+= strlen(b);
			if(b >= buf + 2) {
				if(b-1 > buf && b[-2] == '\\') {
					if(!fgets(buf,BUFSIZE,fp)) {
						return;
						}
					++linenum;
					if(Vflag) printf("%s%4ld: (%2d,%2d) %s%s",CYAN,linenum,ifdefctr,ifdefskip,YELLOW,buf);
					b= buf;
					}
				else break;
				}
			else break;
			}	/* while(1) {...} */
		continue;
		}
	else if(do_asm) {	/* #asm ... #endasm keyword support */
		asmfsa(buf);
		continue;
		}
	else if(Vflag) {
		if(ifdefskip) printf("%s%4ld: (%2d,%2d) %s%s",CYAN,linenum,ifdefctr,ifdefskip,MAGENTA,buf);
		else          printf("%s%4ld: (%2d,%2d) %s%s",CYAN,linenum,ifdefctr,ifdefskip,GREEN,buf);
		}
	b= stpblk(buf);

	for(; *b; b= stpblk(b+1)) {

		if(*b & 0x80) {	/* 8th bit set, not a proper *.c, *.C, or *.h file */
			error(XTDIO_ERROR,"<%s%s%s> has non-ascii characters\n",RVGREEN,filename,GREEN);
			}

		/* handle strings */
		if(*b == '"' || *b == '\'') {
			qCq= b[0] == '"' && b[1] == 'C' && b[2] == '"';
			b  = skipstring(b,buf,fp);
			continue;
			}

		/* handle comments */
		if(b[0] == '/') {
			if(b[1] == '/') break; /* C++ style comment, skip rest of line */
			if(b[1] == '*') b= skipcomment(b+1,buf,fp);
			continue;
			}

		/* run yacc or hdr FSA */
		b= cname(b,cvar);
		if(*b) {
			switch(filemode) {

			case YACC:
				ret= yaccfsa(cvar,b,buf);
				if     (ret == 1) continue;
				else if(ret == 2) goto skipfile;
				break;

			default:
				hdrfsa(cvar,b,buf,ifdefskip);
				break;
				}
			}
		else break;
		}
	}

/* reset FSA */
skipfile:
switch(filemode) {
	case C:			hdrfsa ((char *) NULL,(char *) NULL,(char *) NULL,0);			break;
	case LEX:																		break;
	case TEX:		texfsa(NULL);													break;
	case YACC:		yaccfsa((char *) NULL,(char *) NULL,(char *) NULL);				break;
	case VIM:		vimfsa(NULL);													break;
	case MAPLE:		mplfsa(NULL);													break;
	case MATLAB:	matfsa(NULL);													break;
	default:		error(XTDIO_WARNING,"filemode=%d not supported for reset\n");	break;
	}

}

/* --------------------------------------------------------------------- */
/* cname: this function extracts a C/C++-valid variable name {{{2
 *  Returns: ptr in inp, points to last character in C-valid variable name
 *           cvar : holds null string or a C-valid variable name
 *
 *  A C-valid variable name is:     [_a-zA-Z][_a-zA-Z0-9]
 *  (in regular expression syntax)
 */
static char *cname(char *inp,char *cvar)
{
char *b        = NULL;
char *simpname = NULL;


/* attempt to extract valid C variable name */
inp    = stpblk(inp);
cvar[0]= '\0';

/* Characters permitted in a C variable name include the "$" since some
 * compilers, notably the besotten-with-exceptions VMS o/s's compiler,
 * does.
 */
if(isalpha(*inp) || *inp == '_' || *inp == '$') {
	for(b= simpname= cvar; *inp && (isalpha(*inp) || isdigit(*inp) || *inp == '_' || *inp == '$'); ++b, ++inp) {
		*b= *inp;
		}
	*b= '\0';
	if(pflag) {	/* C++ exception */
		if(inp[0] == ':' && inp[1] == ':') {
			*(b++)   = ':';
			*(b++)   = ':';
			inp     += 2;
			if(inp[0] == '~') {	/* handle destructors, too */
				*(b++)= '~';
				++inp;
				}
			simpname = b;
			for(; *inp && (isalpha(*inp) || isdigit(*inp) || *inp == '_' || *inp == '$'); ++b, ++inp) {
				*b= *inp;
				}
			*b= '\0';
			}
		if(!strcmp(simpname,"operator") &&
		  (
		   (inp[0] == '^' || inp[0] == '=' || inp[0] == '-' || inp[0] == '+' || inp[0] == '/' || inp[0] == '<' || inp[0] == '>' ) &&
		   (inp[1] == '=' || isspace(inp[1]) || inp[1] == '('))) {
			*(b++)= *(inp++);
			if(inp[0] == '=') *(b++)= *(inp++);
			*b= '\0';
			}
		}
	--inp;
	}

return inp;
}

/* --------------------------------------------------------------------- */
/* MaybeFunc: this function handles the may-be-a-function setting {{{2 */
static void MaybeFunc(char *b,int *state)
{

switch(*b)  {
case '(':
	ltl->tagtype = TTmaybefunc;
	*state       = 0;
	if(maybefunc) maybefunc->tagtype= TTglobal;
	maybefunc    = ltl;
	break;
case '*':
	break;
case ';':
case '=':
	if(maybefunc) {
		maybefunc->tagtype = TTglobal;
		maybefunc          = NULL;
		}
	*state= 0;
	break;
default:
	*state= 0;
	break;
	}

}

/* --------------------------------------------------------------------- */
/* asmfsa: this function recognizes asm XREF labels {{{2 */
static void asmfsa(char *buf)
{
Label *match     = NULL;     /* FSA's next state is *match   */
char   cvar[BUFSIZE];
char  *b;
char  *c;
char  *colon;


/* elide ;comment */
b     = strchr(buf,';');
colon = strchr(buf,':');
if(b) *b= '\0';

b= stpblk(buf);

/* XREF label */
if(tolower(b[0]) == 'x' &&
   tolower(b[1]) == 'r' &&
   tolower(b[2]) == 'e' &&
   tolower(b[3]) == 'f' &&
   isspace(b[4])) {
	b= stpblk(b + 4);
	if(*b == '_') {	/* C callable reference */
		for(++b, c= cvar; *b && !isspace(*b); ++b, ++c) *c= *b;
		*c= '\0';
		if(c > cvar) {
			ps_search(hdrbase,cvar,(void **) &match);
			if(!match) makeLabel(cvar,buf,(int) ZeroState,TTasmxref);
			}
		}
	}

/* XDEF label */
else if(tolower(b[0]) == 'x' &&
        tolower(b[1]) == 'd' &&
        tolower(b[2]) == 'e' &&
        tolower(b[3]) == 'f' &&
        isspace(b[4])) {
	b= stpblk(b + 4);
	if(*b == '_') {	/* C callable reference */
		for(++b, c= cvar; *b && !isspace(*b); ++b, ++c) *c= *b;
		*c= '\0';
		if(c > cvar) {
			ps_search(hdrbase,cvar,(void **) &match);
			if(!match) makeLabel(cvar,buf,(int) ZeroState,TTasmxdef);
			}
		}
	else {
		for(c= cvar; *b && !isspace(*b); ++b, ++c) *c= *b;
		*c= '\0';
		if(c > cvar) {
			ps_search(hdrbase,cvar,(void **) &match);
			if(!match) makeLabel(cvar,buf,(int) ZeroState,TTasmxdef);
			}
		}
	}

/* LABEL: */
else if(colon) {
	char *bb;
	b  = stpblk(b);
	bb = b;
	if(*bb == '_') ++b;
	for(b= stpblk(b), c= cvar; *b && b < colon; ++b, ++c) {
		*c= *b;
		if(!isdigit(*b) && !isalpha(*b) && *b != '_') break;
		}
	*c= '\0';
	if(*b == ':' && c > cvar) {
		enum TagType tagtype;
		ps_search(hdrbase,cvar,(void **) &match);
		tagtype  = match? match->tagtype : TTnever;
		if(match && *bb == '_' && match->tagtype == TTasmxdef) {
			freeLabel(match,C);
			match= NULL;
			}
		/* If a label has been xdef'd, its internal to the asm.
		 * Otherwise, it is clear that this label is intended for
		 * use from a C program, and so should be makeLabelled elsewhere.
		 */
		if(!match) makeLabel(cvar,buf,(int) ZeroState,(*bb == '_' && tagtype == TTasmxdef)? TTasmlabelc : TTasmlabel);
		}
	}

}

/* --------------------------------------------------------------------- */
/* forfsa: this function implements the Fortran finite-state automaton {{{2 */
static void forfsa(char *buf)
{

error(XTDIO_ERROR,"forfsa: fortran language processing not implemented yet\n");

}

/* --------------------------------------------------------------------- */
/* hdrfsa: run the Finite State Automaton (c, c++, header files) {{{2
 *
 *  Actions:
 *    D == Discard last Label
 *    M == makeLabel
 *    G == makeLabel if gflag true, nparen==0, and nbrace <= xcbrace
 *
 *  Finite State Automaton Diagram:
 *     punctuation transitions taken from *b
 *     LABELs are contained in cvar
 *     whenever curly brace count drops to zero, xcbrace is set to zero
 *
 *       struct   LABEL     {
 *    0 ------- 1 ----- 2 -----> 0
 *    |             M   |
 *    |                 | LABEL
 *    |                 +------> 0
 *    |                    DG
 *    |
 *    |  enum     LABEL       {        LABEL   ;
 *    +-------- 3 ------- 4 ------- 5 -------+----> 0
 *    |         |   M     |              M   |
 *    |         |         |  ;               +----> 5
 *    |         |         +-------> 0
 *    |         +---> 0   |  D
 *    |                   |
 *    |                   |
 *    |                   | LABEL
 *    |                   +-------> 5
 *    |                      DG
 *    |
 *    | typedef   LABEL     ;
 *    +-------- 6 ----- 7 -----> 0
 *    |             M   |
 *    |                 |   ,        ;
 *    |                 +------- 8 -----> 0
 *    |                 |        |
 *    |                 |        | LABEL
 *    |                 |        +------> 8
 *    |                 |        |   M
 *    |                 |        |
 *    |                 |        |   (       LABEL
 *    |                 |        +------- 9 -------> 9
 *    |                 |                 |    M
 *    |                 |   (             |
 *    |                 +------> 9        |    ;
 *    |                 |   D             +--------> 0
 *    |                 |                 |
 *    |                 | LABEL           |    )          , (==0
 *    |                 +------> 7        +--------- 10 ---------> 8
 *    |                    DM             |           |
 *    |                                   |           |   ;
 *    |                                   |           +----------> 0
 *    |                                   |    ;
 *    | extern "C"                        +---------> 0
 *    +-----------> 0
 *    |  xcbrace=1
 *    |
 *    |  LABEL
 *    +-----------> 0
 */
static void hdrfsa(
  char *cvar,        /* ptr to current variable                             */
  char *b,           /* ptr to last character in current variable in buffer */
  char *buf,         /* current buffered line                               */
  int   ifdefskip)   /* suppports #ifdef .. #else .. #endif                 */
{
char       *ptr        = NULL; /* used to make ps_search happy   */
Label      *match      = NULL; /* FSA's next state is *match     */
static int  nbrace     = 0;    /* qty of open []s                */
static int  ncurly     = 0;    /* qty of open {}s                */
static int  nparen     = 0;    /* qty of open ()s                */
static int  state      = 0;    /* Finite State Automaton's state */
static int  xc_ncurly  = 0;    /* extern "C" exception support   */
static int  classcurly = -2;   /* class..{} level                */


/* reset */
if(!cvar) {

	/* sanity checks */
	if(xc_ncurly != 0 && wflag)
	  error(XTDIO_WARNING,
		"hdrfsa <%s:%ld>: end-of-file but mismatched extern \"C\" curly braces\n",
		fname,linenum);
	if(ncurly != 0 && wflag)
	  error(XTDIO_WARNING,
	    "hdrfsa <%s:%ld>: end-of-file but mismatched curly braces {} !\n",
	    sprt(fname),linenum);
	if(state != ZeroState && state != XState && wflag)
	  error(XTDIO_WARNING,
	    "hdrfsa <%s:%ld>: end-of-file but state(%d) isn't zero!\n",
	    sprt(fname),linenum,state);
	if(nparen != 0 && wflag)
	  error(XTDIO_WARNING,
	    "hdrfsa <%s:%ld>: end-of-file but mismatched parentheses ()!\n",
	    sprt(fname),linenum,state);
	if(nbrace != 0 && wflag)
	  error(XTDIO_WARNING,
	    "hdrfsa <%s:%ld>: end-of-file but mismatched braces [] !\n",
	    sprt(fname),linenum,state);

	qCq       = nbrace = xc_ncurly = ncurly = state = nparen = 0;
	maybefunc = NULL;
	return;
	}

/* keep track of parentheses and braces, but only for the leading sections
 * of if[def|ndef] blocks.
 */
if(!ifdefskip) switch(*b) {

case '(':
	++nparen;
	break;

case ')':
	if(--nparen < 0) {
		if(wflag) error(XTDIO_WARNING,"hdrfsa <%s:%ld>: nparen < 0\n",sprt(fname),linenum);
		nparen= 0;
		}
	break;

case '{':
	/* hdrtag considers open-curly at the beginning of a line to probably
	 * be a beginning of a function.  If the curly brace count > 1 (or 2 if extern "C"...)
	 * (after incrementing the ncurly) and warnings are enabled, a warning
	 * will be issued.  However, the hdrtag program will continue processing
	 * as-is (ie. the possible beginning of function will not be treated as such)
	 */
	++ncurly;
	nparen= 0;
	if(b == buf && ncurly > xc_ncurly + 1 && wflag)
	  error(XTDIO_WARNING,"hdrfsa <%s:%ld>: ncurly=%d but possible beginning-of-function detected\n",sprt(fname),linenum,ncurly);
	break;

case '}':
	if(--ncurly < 0) {
		if(xc_ncurly > 0) ncurly= 0;
		else {
			error(XTDIO_WARNING,"hdrfsa <%s:%ld>: ncurly < 0\n",sprt(fname),linenum);
			ncurly= 0;
			}
		}
	if(ncurly == 0) {
		xc_ncurly= 0;
		}
	if(ncurly < classcurly) classcurly= -2;
	nparen= 0;
	if(b == buf && ncurly > xc_ncurly && wflag)
	  error(XTDIO_WARNING,"hdrfsa <%s:%ld>: ncurly=%d but probable end-of-function detected\n",sprt(fname),linenum,ncurly);
	break;

case '[':
	++nbrace;
	break;

case ']':
	if(--nbrace < 0) nbrace= 0;
	break;

default:
	break;
	}


/* run finite state automaton */
switch(state) {

case MultiZeroState:
	/* eliminate  "label label ... label", except for last label */
	if     (cvar[0])                                         freeLabel(ltl,C);
	else if(ltl->tagtype == TTglobal && ncurly <= xc_ncurly) MaybeFunc(b,&state);
	else                                                     state= ZeroState;
	goto HandleZero;

case MultiXState:
	if(cvar[0]) freeLabel(ltl,C);	/* eliminate  "label label ... label", except for last label */
	else if(ltl->tagtype == TTglobal && ncurly <= xc_ncurly) MaybeFunc(b,&state);
	else        state= XState;
	goto HandleZero;

case ZeroState: /* (matching non-keywords also map to ZeroState) */
case XState:    /* the last cvar was already in the hdrbase      */

HandleZero:
	if(cvar[0]) {
		/* check if label is in hdrbase; if so, match will point to a
		 * Label.  That label's fsa_state will be the next state transition.
		 */
		ptr= ps_search(hdrbase,cvar,(void **) &match);


		/* don't consider globals in a skipped section of a
		 * #ifdef...#else...#endif
		 */
		if(!ifdefskip) {
			/* if match, then label is in hdrbase: it is one of fprior's
			 * ckeys/cppkeys' keywords or has already appeared in the header
			 * file (ie. a duplicate).
			 */
			if(match) {
				if(match->fsa_state == XState && ncurly <= xc_ncurly) {
					/* Don't exclude repeats of label(s):
					 *
					 * Labels are made XState only when:
					 *  gflag is true  (accept globals, prototypes, etc)
					 *  open parenthesis count is zero
					 *  open curly brace count is zero
					 *                         or one  if in an extern "C" { ... }
					 *
					 * The next FSA state is left at its current state
					 */
					makeLabel(cvar,buf,XState,nparen? TTglobalinparen : TTglobal);
					state= (state == ZeroState)? MultiZeroState : MultiXState;
					}
				else {	/* next FSA state given by match->fsa_state */
					state= match->fsa_state;
					break;
					}
				}

			/* label is currently *not* in hdrbase */
			else if(nparen == 0) {
				if(gflag && ncurly <= xc_ncurly) {
					/* Global variables, prototypes, etc, are to be included
					 * in the <hdrtags> file.  Note that cvar is not currently
					 * in hdrbase, so put it on in there.
					 * Also, set state to ZeroState
					 */
					makeLabel(cvar,buf,XState,TTglobal);
					state= (state == ZeroState)? MultiZeroState : MultiXState;
					}
				else if(gflag && pflag && ncurly == classcurly + 1) {
					/* prototype functions declared inside C++ classes and structs */
					makeLabel(cvar,buf,XState,TTmaybefunc);
					state= (state == ZeroState)? MultiZeroState : MultiXState;
					}
				}
			else {
				if(gflag && ncurly <= xc_ncurly && !maybefunc) {
					/* the only labels like this that I want to tag are
					 * functions returning pointers to functions...  COMBAK
					 */
					}
				}
			}
		else if(match) state= match->fsa_state;
		}		/* cvar[0] is alphameric */

	else if(ispunct(*b) && *b != '(' && *b != '*') {
		if(maybefunc && *b == '{') {
			char *colon;
			maybefunc->tagtype = TTfuncname;

			/* generate C++ shortfuncname */
			if(pflag) {
				colon= strchr(maybefunc->label,':');
				if(colon && colon[1] == ':') makeLabel(colon + 2,maybefunc->line,XState,TTshortfuncname);
				}

			/* reset maybefunc */
			maybefunc          = NULL;
			}
		state= ZeroState;
		}
	break;

case ClassLabel:	/* [STRUCT|CLASS] LABEL */
	if(cvar[0]) {
		state= ClassLabelLabel;
		makeLabel(cvar,buf,ZeroState,TTclass);
		}
	else state      = ZeroState;
	break;

case ClassLabelLabel:	/*	[STRUCT|CLASS] LABEL [LABEL|:|{] */
	switch(*b) {

	case '{':	/* [STRUCT|CLASS] LABEL { */
		state      = ZeroState;
		classcurly = ncurly;
		break;

	case ':':
		if(pflag) { /* CLASS LABEL : */
			state= ClassLabelColon;
			break;
			}
		/* deliberate fall through */

	default:	/* STRUCT LABEL LABEL */

		/* free up structure name Label, its not defined here */
		freeLabel(ltl,C);

		/* makeLabel second LABEL if gflag permits */
		if(gflag && cvar[0]) {
			ptr= ps_search(hdrbase,cvar,(void **) &match);
			if(!match && nparen == 0 && ncurly <= xc_ncurly) {
				makeLabel(cvar,buf,ZeroState,TTclasslabel);
				}
			}
		state= ZeroState;
		break;
		}
	break;

case EnumLabel: /* ENUM LABEL */
	if(cvar[0]) {
		state= EnumLabelLabel;
		makeLabel(cvar,buf,ZeroState,TTenumtype);
		}
	else if(*b == '{')              state= EnumLabelCurly;
	else if(*b == ')' || *b == '}') state= ZeroState;
	else                            state= ZeroState;
	break;

case EnumLabelLabel:	/* ENUM LABEL LABEL */
	switch(*b) {

	case ';':	/* ENUM LABEL; */
		freeLabel(ltl,C);
		state= ZeroState;
		break;

	case '{':	/* ENUM LABEL { ...          (}}) */
		state= EnumLabelCurly;	/* makeLabel the internal labels	*/
		break;

	case ')':	/* ENUM LABEL)  (probably end of function header) */
		freeLabel(ltl,C);	/* free up enumeration label */
		state= ZeroState;
		break;

	case '}':	/* ENUM LABEL} */
		state= ZeroState;
		break;

	default:	/* ENUM LABEL LABEL */
		freeLabel(ltl,C);	/* free up enumeration label */

		if(gflag && cvar[0]) {
			ptr= ps_search(hdrbase,cvar,(void **) &match);
			if(!match && nparen == 0 && ncurly <= xc_ncurly) {
				makeLabel(cvar,buf,ZeroState,TTenumtype);
				}
			}
		state= EnumLabelCurly;
		break;
		}
	break;

case EnumLabelCurly:	/* ENUM LABEL [{] LABEL [, LABEL ...] [}] [;] */
	if(cvar[0]) {
		makeLabel(cvar,buf,ZeroState,TTenumlabel);
		}
	else if(*b == ';') state= ZeroState;
	else if(*b == '}') state= ZeroState;
	else if(*b == '=') state= EnumLabelResume;
	else if(*b == ')') state= ZeroState;
	else if(*b == '(') {	/* probably a enum-returning function or prototype */
		state= ZeroState;
		if(gflag && ltl) {
			ltl->fsa_state= XState;
			ltl->tagtype  = TTglobal;
			}
		}
	break;

case EnumLabelResume:
	if     (*b == ',') state= EnumLabelCurly;
	else if(*b == ';') state= ZeroState;
	else if(*b == '}') state= ZeroState;
	break;

case TypedefLabel: /* TYPEDEF LABEL */
	if(cvar[0]) {
		ptr= ps_search(hdrbase,cvar,(void **) &match);
		if(!match || match->tagtype != TTnever) {
			makeLabel(cvar,buf,ZeroState,TTtypedef);
			state= TypedefLabelLabel;
			}
		else if(match->fsa_state == EnumLabel) state= EnumLabel;	/* typedef enum ... */
		}
	else if(*b == ';') state= ZeroState;
	else if(*b == '{') state= TypedefLabelCurlyStop;
	else if(*b == '[') state= TypedefLabelBrace;
	else if(*b == '(') state= TypedefLabelParenStart;
	else               state= ZeroState;
	break;

case TypedefLabelLabel:
	if(cvar[0]) { /* TYPEDEF LABEL LABEL... */
		freeLabel(ltl,C);
		makeLabel(cvar,buf,ZeroState,TTtypedef);
		}
	else if(*b == ';') state= ZeroState;
	else if(*b == ',') state= TypedefLabelListStart;
	else if(*b == '{') state= TypedefLabelCurlyStop;	/*}*/
	else if(*b == '(') {
		freeLabel(ltl,C);
		state= TypedefLabelParenStart;
		}
	else if(*b == '[') state= TypedefLabelBrace;
	break;

case TypedefLabelListStart:
	if(cvar[0]) {	/* TYPEDEF LABEL [LABEL ...] , LABEL ... */
		makeLabel(cvar,buf,ZeroState,TTtypedef);
		}
	else if(*b == ';') state= ZeroState;
	else if(*b == '(') state= TypedefLabelParenStart;
	else if(*b == '[') state= TypedefLabelBrace;
	else if(*b == '{') state= TypedefLabelCurlyStop;
	else if(*b == ')') {
		if(ifdefskip) state= ZeroState;
		else          error(XTDIO_ERROR,"hdrfsa <%s:%ld>: nparen < 0\n",sprt(fname),linenum);
		}
	break;

case TypedefLabelParenStart:
	if(cvar[0]) {	/* TYPEDEF LABEL ... (...LABEL */
		makeLabel(cvar,buf,ZeroState,TTtypedef);
		}
	else if(*b == ')') state= TypedefLabelParenStop;
	else if(*b == ';') state= ZeroState;
	else if(*b == '{') state= TypedefLabelCurlyStop;
	break;

case TypedefLabelParenStop:	/* TYPEDEF LABEL ... (... LABEL ) ... */
	if     (*b == ',' && nparen == 0) state= TypedefLabelListStart;
	else if(*b == ';')                state= ZeroState;
	break;

case Extern:	/* EXTERN [LABEL|"C" {]        }*/
	if(*b == '{') {
		for(--b; b > buf+2; --b) if(!isspace(*b)) break;
		if(qCq) {	/* special exception for extern "C" {              }*/
			xc_ncurly= 1;
			state    = ZeroState;
			}
		}
	else if(ncurly <= xc_ncurly) {
		state  = ExternLabelStuff;
		goto extrnlbl;
		}
	else state=  ZeroState;
	break;

case ClassLabelColon:	/* CLASS LABEL : ... { */			/*}*/
	if(*b == '{') state= ZeroState;
	break;

case TypedefLabelCurlyStop:	/* TYPEDEF LABEL ... { ... } */
	if(ncurly == xc_ncurly) state= TypedefLabelCurly;
	break;

case ExternLabelStuff:	/* EXTERN LABEL [;|,|LABEL] ... ; */
extrnlbl:
	if(nparen == 0) {
		if(cvar[0]) {	/* EXTERN ... LABEL ... */
			/* check if label is in hdrbase; if so, match will point to a
			 * Label.  That label's fsa_state will be the next state transition.
			 */
			ptr= ps_search(hdrbase,cvar,(void **) &match);

			/* no point putting in multiple extern declarations; the one that's
			 * in there will do nicely. (also avoids "struct")
			 */
			if(!match) {
				makeLabel(cvar,buf,XState,TTextern);
				}
			}
		else if(*b == ';') state= ZeroState;
		}
	break;

case TypedefLabelCurly: 	/* typedef ... label {,label...}; */
	if(cvar[0]) {
		makeLabel(cvar,buf,ZeroState,TTtypedef);
		}
	else if(*b == ';') state= ZeroState;
	break;

case TypedefLabelBrace:	/* TYPEDEF LABEL ... [... LABEL ] ... */
	if     (*b == ',' && nbrace == 0) state= TypedefLabelListStart;
	else if(*b == ';')                state= ZeroState;
	break;

default:
	if(wflag) error(XTDIO_WARNING,"hdrfsa <%s:%ld>: bad state=%d\n",sprt(fname),linenum,state);
	break;
	}

/* "C" indicator always set to false at end of hdrfsa.  That way, only a state
 * immediately following the "C" will be affected by qCq.
 */
qCq= 0;

}

/* --------------------------------------------------------------------- */
/* matprepro: this function preprocesses to load multi-line commands {{{2 */
static char *matprepro(char *buf,FILE *fp)
{
char *b;
int   bufsize;
int   buflen;


srmtrblk(buf);
buflen  = strlen(buf);
b       = buf + buflen;
bufsize = BUFSIZE;
while(b[0] == '.' && b[-1] == '.' && b[-2] == '.') {
	b[-2]    = '\0';
	bufsize -= buflen;
	fgets(b,bufsize,fp);
	srmtrblk(buf);
	buflen = strlen(buf);
	b     += buflen;
	}

return buf;
}

/* --------------------------------------------------------------------- */
/* matfsa: this function {{{2 */
static void matfsa(char *buf)
{
char  *b             = NULL;
char  *bb            = NULL;
char  *dquote        = NULL;
Label *state         = NULL;
char   word[MAXWORD];


/* ignore reset */
if(!buf) {
	return;
	}

/* elide vim comments -- not really foolproof, but nothing after
 * any strings appear are something I want, either.
 */
dquote = strchr(buf,'%');
if(dquote) *dquote= '\0';

b  = stpblk(buf);
if(!*b) {
	return;
	}
bb= ps_search(matbase,b,(void **) &state);

if(state) {
	switch(state->fsa_state) {

	case matZeroState:
		break;

	case matFunction:
		bb= stpblk(bb);
		if(*bb == '[') { /* skip output argument list */
			bb= strchr(bb,']');
			if(bb) bb= stpblk(bb+1);
			}
		if(*bb == '=') bb= stpblk(bb+1);
		if(isalpha(*bb)) {
			char *w;
			for(w= word; *bb && (isalnum(*bb) || *bb == '_'); ++bb,++w) *w= *bb;
			*w= '\0';
			makeLabel(word,buf,matFunction,TTmatfunc);
			}
		break;

	default:
		break;
		}
	}

}

/* --------------------------------------------------------------------- */
/* mplfsa: TeX/LaTeX automaton for recognizing labels {{{2 */
static void mplfsa(char *buf)
{
char  *b;
char  *bb;
char  *colon;
char  *w;
char   word[MAXWORD];


/* ignore reset */
if(!buf) {
	return;
	}

/* elide comments */
for(b= buf; *b; ++b) {
	if     (b[0] == '\\' && b[1]) ++b;
	else if(*b == '#')            break;
	}
*b= '\0';

/* extract variable name := ... */
word[0] = '\0';
colon   = strchr(buf,':');
if(colon && colon[1] == '=') {
	for(b= colon - 1; b >= buf; --b) if(!isspace(*b)) break;
	if(b >= buf) {
		for(bb= b; bb >= buf; --bb) if(isspace(*bb)) break;
		++bb;
		for(w= word; bb <= b; ++bb, ++w) {
			if(!isalnum(*bb) && *bb != '_') {
				return;
				}
			*w= *bb;
			}
		*w= '\0';
		}

	if(!word[0]) {
		return;
		}

	b= stpblk(colon + 2);
	if(b[0] == 'p' && b[1] == 'r' && b[2] == 'o' && b[3] == 'c') {
		b= stpblk(b+4);
		if(*b == '(') {	/* name := proc(... */
			makeLabel(word,buf,MplProc,TTmplproc);
			}
		}
	else {
		while(*b && (isspace(*b) || *b == '(' || *b == ')' || *b == ',' || isalnum(*b))) ++b;
		if(b[0] == '-' && b[1] == '>') {	/* name := [(...,...)] -> */
			makeLabel(word,buf,MplMap,TTmplmap);
			}
		else if(gflag) {	/* name := */
			makeLabel(word,buf,MplVar,TTmplvar);
			}
		}
	}

}

/* --------------------------------------------------------------------- */
/* texfsa: TeX/LaTeX automaton for recognizing labels {{{2 */
static void texfsa(char *buf)
{
char  *b;
char  *bb;
char  *bend;
Label *state= NULL;


/* ignore reset */
if(!buf) {
	return;
	}

for(b= buf; *b; ++b) {
	if     (b[0] == '\\' && b[1]) ++b;
	else if(*b == '%')            break;
	}
bend= b;

/* look for \label{...} and \cite{...} */
for(b= strchr(buf,'\\'); b && b < bend; b= strchr(b+1,'\\')) {
	bb= ps_search(texbase,b+1,(void **) &state);
	if(state && (*bb == '{' || *bb == '[')) {
		char  label[BUFSIZE];
		char *lbl;
		/* ok a \label{...}, \cite{...}, etc found; set up the ... as a label */
		bb= stpblk(bb+1);
		if(*bb == '\\') ++bb;
		for(lbl= label; *bb && (*bb != '}' && *bb != ']') && !isspace(*bb); ++lbl, ++bb) *lbl= *bb;
		*lbl= '\0';
		if(lbl > label) {
			switch(state->fsa_state) {
				/* \bibitem[...]... */
				case texBibitem: makeLabel(label,buf,state->fsa_state,TTtexbibitem); break;
                /* \cite{...}  */
				case texCite:
					/* further break label into comma-delimited fields (if any commas) */
					if(strchr(label,',')) {
						for(lbl= strtok(label,","); lbl; lbl= strtok(NULL,",")) {
							makeLabel(lbl,buf,state->fsa_state,TTtexcite);
							}
						}
					else makeLabel(label,buf,state->fsa_state,TTtexcite);
					break;
				case texCommand: makeLabel(label,buf,state->fsa_state,TTtexcommand); break;
                /* \newcommand{\...}     */
                /* \renewcommand{\...}   */
                /* \providecommand{\...} */
				case texEnvironment: makeLabel(label,buf,state->fsa_state,TTtexenvironment); break;
				/* \Function{...} */
				case texFunction: makeLabel(label,buf,state->fsa_state,TTtexfunction); break;
                /* \label{...} */
				case texLabel:   makeLabel(label,buf,state->fsa_state,TTtexlabel); break;
				default: break;
				}
			}
		}
	}

}

/* --------------------------------------------------------------------- */
/* vimfsa: this function runs the vim finite state automaton {{{2 */
static void vimfsa(char *buf)
{
char       label[BUFSIZE];
char      *lbl    = NULL;
char      *b      = NULL;
char      *bb     = NULL;
char      *dquote = NULL;
Label     *state  = NULL;


/* ignore reset */
if(!buf) {
	return;
	}

/* elide vim comments -- not really foolproof, but nothing after
 * any strings appear are something I want, either.
 */
dquote = strchr(buf,'"');
if(dquote) *dquote= '\0';

b      = stpblk(buf);
bb     = ps_search(vimbase,b,(void **) &state);

if(state) {
	switch(state->fsa_state) {

	case VimSyntax:
		bb= ps_search(vimbase,stpblk(bb+1),(void **) &state);

		if(state) {
			switch(state->fsa_state) {
			case VimMatch:
			case VimRegion:
			case VimCluster:
			case VimKeyword:
				bb= stpblk(bb+1);
				for(lbl= label; *bb && !isspace(*bb); ++lbl, ++bb) *lbl= *bb;
				*lbl= '\0';
				if(dquote) *dquote= '"';
				if(lbl > label) makeLabel(label,buf,state->fsa_state,TTvimsyntax);
				break;

			default:
				break;
				}
			}
		break;

	case VimFunction:
		if(*bb == '!') ++bb;
		bb= stpblk(bb+1);
		for(lbl= label; *bb && !isspace(*bb) && *bb != '('; ++lbl, ++bb) *lbl= *bb;
		*lbl= '\0';
		if(dquote) *dquote= '"';

		/* generate extra (possibly duplicate) labels:
		 *   <SID>func -> func
		 *   s:func    -> func
		 * _don't_ generate <SID>func and s:func labels -- tags doesn't grok that
		 */
		if(label[0] == '<' && (label[1] == 's' || label[1] == 'S') &&
		                      (label[2] == 'i' || label[2] == 'I') &&
		                      (label[3] == 'd' || label[3] == 'D') &&
		   label[4] == '>') makeLabel(label+5,buf,VimFuncName,TTvimfuncshort);
		else if(label[0] == 's' && label[1] == ':')  makeLabel(label+2,buf,VimFuncName,TTvimfuncshort);
		else if(lbl > label) makeLabel(label,buf,VimFuncName,TTvimfunc);
		break;

	case VimCommand:
		if(*bb == '!') ++bb;
		for(bb= stpblk(bb); *bb && *bb == '-'; bb= stpnxt(bb,"%s%b")) ;
		if(*bb && isalpha(*bb)) {
			sscanf(bb,"%s",label);
			makeLabel(label,buf,VimCmdName,TTvimcmd);
			}
		break;

#ifdef USE_VIMMAP
	case VimMap:
		if(*bb == '!') ++bb;
		for(bb= stpblk(bb); bb && *bb && *bb == '<'; bb= stpblk(++bb)) {
			sscanf(bb,"%s",label);
			for(lbl= label; *lbl; ++lbl) if(isupper(*lbl)) *lbl= tolower(*lbl);
			if(!strcmp(label,"<leftmouse>")   ||
			   !strcmp(label,"<rightmouse>")  ||
			   !strcmp(label,"<cr>")          ||
			   !strcmp(label,"<tab>")         ||
			   !strcmp(label,"<space>")       ||
			   !strcmp(label,"<insert>")      ||
			   !strcmp(label,"<home>")        ||
			   !strcmp(label,"<pgup>")        ||
			   !strcmp(label,"<pgdn>")        ||
			   !strcmp(label,"<del>")         ||
			   !strcmp(label,"<end>")         ||
			   !strcmp(label,"<left>")        ||
			   !strcmp(label,"<right>")       ||
			   !strcmp(label,"<up>")          ||
			   !strcmp(label,"<down>")        ||
			   (label[1] == 'c' && label[2] == '-') ||
			   (label[1] == 'm' && label[2] == '-') ||
			   (label[1] == 's' && label[2] == '-') ||
			   (label[1] == 'l' && label[2] == 'e' && label[3] == 'a' && label[4] == 'd' && label[5] == 'e' && label[6] == 'r') ||
			   (label[1] == 'f' && isdigit(label[2])) ) break;
			bb= strchr(bb,'>');
			}
		if(bb && *bb) {
			sscanf(bb,"%s",label);
			makeLabel(label,buf,VimCmdName,TTvimcmd);
			}
		break;
#endif	/* USE_VIMMAP */

	default:
		break;
		}
	}

}

/* --------------------------------------------------------------------- */
/* yaccfsa: run yacc-oriented finite state automaton {{{2 
 *  Returns: 0=continue normally
 *           1=skip rest of line
 *           2=skip rest of file
 */
static int yaccfsa(char *cvar,char *b,char *buf)
{
Label           *match       = NULL;     /* FSA's next state is *match     */
static  int      ncurly      = 0;        /* nesting {} level               */
static  char     ppseparator = 0;        /* counter of '%%' separators     */
static  int      state       = 0;        /* Finite State Automaton's state */
enum    TagType  recordtype  = TTnever;  /* record label flag              */


/* handle reset */
if(!cvar) {
	state      = (int) ZeroState;
	ppseparator= 0;
	ncurly     = 0;
	return 0;
	}

/* keep track of curlies, and reset state upon each new line
 * for the first yacc section
 */
if     (*b == '{') ++ncurly;
else if(*b == '}') --ncurly;
if(b == buf && ppseparator == 0) state= ZeroState;

switch(state) {

case ZeroState:	/* yacc parsing state */
	if(b == buf && *b == '%') {
		switch(buf[1]) {
		case '%':
			if(++ppseparator == 1) state= YaccLabel;
			return ppseparator;

		case '{':
			state= YaccPercentCurly;
			break;

		default:
			state= YaccKeyWord;
			break;
			}
		}
	break;

case YaccKeyWord:	/* %? */
	if(cvar[0]) {
		(void) ps_search(yaccbase,cvar,(void **) &match);
		state= (match && match->fsa_state != YaccXState)? match->fsa_state : ZeroState;
		}
	else state= ZeroState;
	break;

case YaccLeft:	/* %left ... */
	recordtype= TTyaccleft;
	break;

case YaccRight:	/* %right ... */
	recordtype= TTyaccright;
	break;

case YaccToken:	/* %token	*/
	state= YaccTokenName;
	break;

case YaccTokenName:	/* %token <...> */
	if(*b == '>') state= YaccTokenList;
	break;

case YaccTokenList:	/* %token <...> ... */
	recordtype= TTyacctoken;
	break;

case YaccType:	/* %type	*/
	if(*b == '<') state= YaccTypeName;
	else          state= ZeroState;
	break;

case YaccTypeName: /* %type <...> */
	if(*b == '>') state= YaccTypeList;
	break;

case YaccTypeList: /* %type <...> ... */
	recordtype= TTyacctype;
	break;

case YaccUnion:	/* %union	*/
	/* This code flags previously unseen labels as TTyaccunionmaybe;
	 * when a comma or semi-colon is seen, the previous label, *if* it was TTyaccunionmaybe,
	 * is promoted to TTyaccunion.
	 */
	if(ncurly == 0) state= ZeroState;
	else if(cvar[0]) {
		if(ltl && ltl->tagtype == TTyaccunionmaybe) freeLabel(ltl,YACC);
		(void) ps_search(yaccbase,cvar,(void **) &match);
		if(!match) makeLabel(cvar,buf,YaccXState,TTyaccunionmaybe);
		}
	else if(*b == ',' || *b == ';') {
		if(ltl && ltl->tagtype == TTyaccunionmaybe) ltl->tagtype= TTyaccunion;
		}
	break;

case YaccPercentCurly:	/* "%{" */
	if(ncurly == 0) state= ZeroState;
	if(cvar[0]) {
		(void) ps_search(hdrbase,cvar,(void **) &match);
		if(!match) {
			(void) ps_search(yaccbase,cvar,(void **) &match);
			if(!match) makeLabel(cvar,buf,YaccXState,TTyaccvar);
			}
		}
	break;

case YaccLabel:	/* label */
	if(*b == ':') state     = YaccStmts;
	else          recordtype= TTyacclabel;
	break;

case YaccStmts: /* label : statements... */
	if(ncurly == 0 && *b == ';') state= YaccLabel;
	break;

default:
	if(wflag) error(XTDIO_WARNING,"yaccfsa <%s:%ld>: bad state=%d\n",sprt(fname),linenum,state);
	break;
	}

/* record label if yaccfsa so requests */
if(recordtype != TTnever && cvar[0]) {
	(void) ps_search(yaccbase,cvar,(void **) &match);
	if(match) {
		switch(match->tagtype) {

		case TTnever:
			state= match->fsa_state;
			break;

		case TTyacctype:
			makeLabel(cvar,buf,YaccXState,recordtype);
			break;

		default:
			/* leave state unchanged */
			break;
			}
		}
	else makeLabel(cvar,buf,YaccXState,recordtype);
	}

return 0;
}

/* --------------------------------------------------------------------- */
/* makeLabel: makes Labels {{{2 */
static void makeLabel(
  char         *label,
  char         *line,
  int           fsa_state,
  unsigned char tagtype)
{
char  *paren;
char  *lb;
char   locbuf[BUFSIZE];
int    slen;
Label *match = NULL;	/* FSA's next state is *match		*/

#ifdef __GNUC__
#else
#endif

lb= stpblk(label);
if(!*lb) {	/* skip blank/all-white-space labels */
	if(wflag) error(XTDIO_WARNING,"makeLabel given a null label at line%d<%s>\n",linenum,sprt(line));
	return;
	}

/* allocate and double-link */
double_link(Label,lhd,ltl,"makeLabel");

/* easy Label initialization */
stralloc(ltl->label,lb,"makeLabel init");
ltl->fname     = fname;
ltl->fsa_state = fsa_state;
ltl->tagtype   = (enum TagType) tagtype;
ltl->isstat    = sflag;

/* insert the Label into the matchbase if:
 *   the label hasn't been seen before, or
 *   if the new Label has higher priority
 */
(void) ps_search(matchbase,ltl->label,(void **) &match);
if(match) {
	/* PRMASK restricts priority comparison to the not-restricted bits */
	if((PRMASK & priority[match->tagtype]) > (PRMASK & priority[tagtype])) {
		ps_insert(matchbase,lb,ltl);
		}
	}
else {
	ps_insert(matchbase,lb,ltl);
	}

if(line) {	/* The label is *not* a "struct" "typedef" etc initialization label */
	/* generate ltl->label */
	if(eflag) {	/* use ETags style */
		char hold= 0;
		paren= strpbrk(line,")");
		if(!paren) paren= strpbrk(line,"(=;{[");	/* ]})" */
		if(paren) {
			hold    = paren[1];
			paren[1]= '\0';
			}
		else srmtrblk(line);
		sprintf(locbuf,"%s%c%s%c%ld,%ld\n",
		  line,
		  127,
		  label,
		  1,
		  linenum,
		  tagtell);
		if(paren) paren[1]= hold;
		}

	/* sflag true: make the "search string" a linenum only */
	else if(sflag) sprintf(locbuf,"%ld",linenum);

	/* normal hdrtag label making */
	else {
		strcpy(locbuf,line);
		slen= strlen(locbuf);
		if(slen > 0 && locbuf[slen-1] == '\n') locbuf[slen-1]= '\0';
		}
	stralloc(ltl->line,locbuf,"makeLabel line");
	}
else ltl->line= NULL;


}

/* --------------------------------------------------------------------- */
/* freeLabel: this function frees memory used by and de-links a Label {{{2 */
static void freeLabel(
  Label            *label,
  enum LanguageMode filemode)
{
PS *base;

if(label) {
	if(label->tagtype != TTnever) {

		/* prevent attempted use of a label which has been deleted.
		 * This is why maybefunc was made global; otherwise, I'd've
		 * left it as static in hdrfsa().
		 */
		if(maybefunc == label) maybefunc= NULL;

		/* remove word from database */
		base= (filemode == YACC)?   yaccbase :
		      (filemode == TEX)?    texbase  :
		      (filemode == C)?      hdrbase  :
		      (filemode == VIM)?    vimbase  :
		      (filemode == MAPLE)?  mplbase  : matchbase;
		ps_delete(base,label->label,(void *) NULL);

		if(label->label) free(label->label);
		if(label->line)  free(label->line);
		label->label     = NULL;
		label->line      = NULL;
		label->fname     = NULL;
		label->tagtype   = TTnever;
		label->fsa_state = 0;
		delete_double_link(Label,label,lhd,ltl);

		}
	}
else if(wflag) error(XTDIO_WARNING,"freeLabel: was given a null pointer to free\n");
}

/* --------------------------------------------------------------------- */
/* skipstring: returns a pointer to end-of-string {{{2 */
static char *skipstring(char *b,char *buf,FILE *fp)
{
char strchar;


strchar= *(b++);
do {
	for(; *b; ++b) {
		if(*b == '\\') ++b;
		else if(*b == strchar) {
			return b;
			}
		}
	if(b[-1] == '\n' && b[-2] != '\\') {	/* forcibly end the string if at end-of-line without \ */
		b= buf;
		++linenum;
		if(!fgets(buf,BUFSIZE,fp)) buf[0]= '\0';
		return b;
		}
	b= buf;
	++linenum;
	} while(fgets(buf,BUFSIZE,fp));

error(XTDIO_ERROR,"skipstring<%s>: end-of-file while in string\n",sprt(fname));

return (char *) NULL;
}

/* --------------------------------------------------------------------- */
/* skipcomment: returns a pointer to end-of-comment {{{2 */
static char *skipcomment(char *b,char *buf,FILE *fp)
{

++b;	/* get past comment's beginning asterisk */

do {
	for(; *b; ++b) if(b[0] == '*' && b[1] == '/') {
		return b+1;
		}
	b= buf;
	++linenum;
	} while(fgets(buf,BUFSIZE,fp));

error(XTDIO_ERROR,"skipcomment<%s>: end-of-file while inside a comment\n",sprt(fname));

return (char *) NULL;
}

/* --------------------------------------------------------------------- */
/* hdrtagout: write <hdrtags> file {{{2 */
static void hdrtagout(void)
{
char    tagbuf[BUFSIZE];
long    itag            = 0;
long    jtag            = 0;
long    lblcnt          = 0;
long    xcnt            = 0;
long    restricted_cnt  = 0;
FILE   *fpout           = NULL;
Label  *label           = NULL;
Label **tagbase         = NULL;


if(!lhd) {
	error(XTDIO_NOTE,"hdrtag: no tags present\n");
	return;
	}

/* count qty of non-restricted labels */
for(label= lhd, lblcnt= 0; label; label= label->nxt) {
	if(priority[label->tagtype] & PRESTRICT) ++restricted_cnt;
	++lblcnt;
	}

lblcnt-= (fsa_lblcnt + restricted_cnt);

/* read in <tags> file */
if(mflag) {
	/* verbose mode support */
	if(vflag) fprintf(stderr,"%smerging hdrtag output with <%stags%s>\n",CYAN, RVGREEN,GREEN);
	lblcnt+= mergetags();
	}

/* check if any labels were made */
if(lblcnt <= 0) {
	if(wflag)    error(XTDIO_WARNING,"no labels found\n");
	if(priority) free((char *) priority);
	exit(0);
	}

/* allocate tagbase, a linear vector of strings (to hold tag lines) */
tagbase= (Label **) calloc((size_t) lblcnt,sizeof(Label *));
outofmem(tagbase,"hdrtagout: unable to allocate %d Label*\n",lblcnt);

/* verbose mode support */
if(vflag) {
	if(mflag) fprintf(stderr,"%sgenerating %s%ld%s more tags in <%stags%s>\n",
	  CYAN,GREEN,lblcnt,
	  CYAN,RVGREEN,GREEN);
	else fprintf(stderr,"%sgenerating %s%ld%s tags in <%shdrtags%s>\n",
	  CYAN,GREEN,lblcnt,
	  CYAN,RVGREEN,GREEN);
	}

/* load tagbase, generate complete Label lines, and free old Label lines */
if(sflag) for(itag= 0, label= lhd; label; label= label->nxt) {
	if(priority[label->tagtype] & PRESTRICT) continue;

	/* short tags */
	if(priority[label->tagtype] != PNEVER) {
		sprintf(tagbuf,"%s\t%s\t%s\n",
		  label->label,
		  label->fname,
		  label->line);
		free((char *) label->line);
		stralloc(label->line,tagbuf,"hdrtagout: while generating tagbase");


		if(itag >= lblcnt) error(XTDIO_ERROR,"hdrtagout: itag=%d but lblcnt=%d!\n", itag,lblcnt);
		tagbase[itag++]= label;
		}
	}

else for(itag= 0, label= lhd; label; label= label->nxt) {
	if(priority[label->tagtype] & PRESTRICT) continue;

	/* usual tags */
	if(priority[label->tagtype] != PNEVER) {
		if(label->isstat) sprintf(tagbuf,"%s\t%s\t%s%s",
		  label->label,
		  label->fname,
		  label->line,
		  (!sflag && tflag)? "" : "\n");
		else sprintf(tagbuf,"%s\t%s\t/^%s/%s",
		  label->label,
		  label->fname,
		  aflag? antimagic(label->line) : label->line,
		  (!sflag && tflag)? "" : "\n");
		free((char *) label->line);
		stralloc(label->line,tagbuf,"hdrtagout: while generating tagbase");


		if(itag >= lblcnt) error(XTDIO_ERROR,"hdrtagout: itag=%d but lblcnt=%d!\n", itag,lblcnt);
		tagbase[itag++]= label;
		}
	}

/* insure that lblcnt is positive */
lblcnt= itag;
if(lblcnt <= 0) {
	if(wflag) error(XTDIO_WARNING,"hdrtag unable to find any labels!\n");
	free((char *) tagbase);
	if(vflag) fprintf(stderr,"%sdone%s\n",CYAN,NRML);
	return;
	}

/* sort the tag strings */
if(vflag) fprintf(stderr,"%ssorting %s%ld%s tags\n",CYAN,GREEN,lblcnt,CYAN);
#ifdef __PROTOTYPE__
qsort(tagbase,
      (size_t) lblcnt,
      sizeof(Label *),
      (int (*)(const void *,const void *)) tag_compare);
#else
qsort(tagbase,
      (size_t) lblcnt,
      sizeof(Label *),
      (int (*)()) tag_compare);
#endif

/* write out the tags */
if(vflag) fprintf(stderr,"%swriting <%stags%s>\n",
  CYAN,
  RVGREEN,
  GREEN);

/* open up the output tagfile (normally <tags>) */
if(!strcmp(tagfile,"-")) fpout= stdout;
else {
#ifdef vms
	fpout= fopen(tagfile."w");
#else
	fpout= fopen(tagfile,"w");
#endif
	}

if(!fpout) {
	error(XTDIO_ERROR,"unable to write <%stags%s>\n", RVGREEN, GREEN);
	}

if(xflag) {	/* exclude duplicates */

	for(itag= jtag= xcnt= 0, xcnt= 0; itag < lblcnt; itag= jtag) {
		PrintTag(tagbase[itag],fpout);

		/* after sorting, highest priority (lowest priority[tagtype]) is
		 * tagbase[itag].  Subsequent tags with identical labels are
		 * of same or lower priority.  Will include only those tags
		 * having same or higher priority than xflag.
		 */
		for(jtag= itag + 1; jtag < lblcnt; ++jtag) {
			if(strcmp(tagbase[itag]->label,tagbase[jtag]->label)) break;
			if((PRMASK & priority[tagbase[jtag]->tagtype]) <= xflag) {
				PrintTag(tagbase[jtag],fpout);
				continue;
				}
			++xcnt;

			freeLabel(tagbase[jtag],C);
			tagbase[jtag] = NULL;
			}
		freeLabel(tagbase[itag],C);
		tagbase[itag] = NULL;
		}	/* for itag < lblcnt loop */

	/* report on qty of excluded tags */
	if(vflag && xcnt > 0) fprintf(stderr,"%sExcluded %s%ld%s duplicate tags\n",CYAN,GREEN,xcnt,CYAN);
	}
else {	/* include all tags */
	for(itag= 0; itag < lblcnt; ++itag) {	/* print out tags */
		PrintTag(tagbase[itag],fpout);

		freeLabel(tagbase[itag],C);
		tagbase[itag]= NULL;
		}
	}

fclose(fpout); /* close file handle */
free((char *) tagbase);
if(vflag) fprintf(stderr,"%sdone%s\n",CYAN,NRML);

}

/* --------------------------------------------------------------------- */
/* PrintTag: this function {{{2 */
static void PrintTag(
  Label *tag,  
  FILE  *fpout)
{

if(zflag) fprintf(fpout,"%s!%03d ",tag->label,TTend - tag->tagtype);
if(tflag && !tag->isstat) {

	if(Xflag)
	  fprintf(fpout,"%s;\"\t%c\tfile:\n",
	  tag->line,
	  (0 <= tag->tagtype && tag->tagtype < TTend)?  Ltagtype[tag->tagtype][3] : '?');

	/* include tagtype field */
	else if(tflag == 1)
	  fprintf(fpout,"%s;\"%s\n",
	  tag->line,
	  (0 <= tag->tagtype && tag->tagtype < TTend)?  Ltagtype[tag->tagtype] + 2 : "???");

	/* include tagtype field and priority */
	else if(tflag == 2)
	  fprintf(fpout,"%s;\"%s %d\n",
	  tag->line,
	  (0 <= tag->tagtype && tag->tagtype < TTend)?  Ltagtype[tag->tagtype] + 2 : "???",
	  (PRMASK & priority[tag->tagtype]));
	}
else fputs(tag->line,fpout);


}

/* ------------------------------------------------------------------------ */
/* hdretagout: generates etags style output (TAGS) {{{2 */
static void hdretagout(void)
{
unsigned  bytecnt;
Label    *lbl;
Label    *ilbl;
FILE     *fpout;


if(!lhd) {
	error(XTDIO_NOTE,"hdrtag: no tags present\n");
	return;
	}

#ifdef vms
fpout= fopen("TAGS","w");
#else
fpout= fopen("TAGS","w");
#endif
if(!fpout) error(XTDIO_ERROR,"unable to write <%sTAGS%s>\n",RVGREEN,GREEN);

for(lbl= lhd; lbl; lbl= lbl->nxt) if((PRMASK & priority[lbl->tagtype]) != PNEVER) {

	/* write out filename */
	if(lbl == lhd || lbl->fname != lbl->prv->fname) {
		if(!lbl->fname) error(XTDIO_SEVERE,"hdretagout: lbl->fname is null for the case <%s>\n",
		  (lbl == lhd)? "lbl == lhd" : "lbl->fname != lbl->prv->fname");

		/* lookahead: etags (TAGS) files contain qty of bytes which follow this line
		 * that are associated with this file.  Pain in the Aft.
		 */
		for(ilbl= lbl, bytecnt= 0; ilbl && lbl->fname == ilbl->fname; ilbl= ilbl->nxt) {
			bytecnt+= strlen(ilbl->line);
			}
		fprintf(fpout,"%c\n%s,%d\n",12,lbl->fname,bytecnt);
		}

	/* write out labels */
	if(lbl->line) {
		fputs(lbl->line,fpout);
		}
	else error(XTDIO_SEVERE,"hdretagout: lbl->line is null\n");
	}

/* output file */
fclose(fpout);

}

/* --------------------------------------------------------------------- */
/* tag_compare: compare two tags (uses label and then priority) {{{2 
 *  Returns <0: tag1 < tag2
 *          =0: tag1 = tag2
 *          >0: tag1 > tag2
 */
static int tag_compare(const Label **tag1,const Label **tag2)
{
int ret;
int tagpriority1;
int tagpriority2;

tagpriority1= PRMASK & priority[(*tag1)->tagtype];
tagpriority2= PRMASK & priority[(*tag2)->tagtype];

ret= strcmp((*tag1)->label,(*tag2)->label);
if(!ret) {	/* same label string, so compare priorities */
	ret= tagpriority1 - tagpriority2;
	}

return ret;
}

/* --------------------------------------------------------------------- */
/* antimagic: this function converts a line with potentially "magic" {{{2 
 *  characters into one with those characters escaped.  It also refuses
 *  to allow comments into the search string.
 */
static char *antimagic(char *line)
{
char       *a;
char       *b;
static char buf[BUFSIZE+BUFSIZE+1];
char       *l;


for(b= buf, l= line; *l; ++l, ++b) {
	if(aflag) {	/* apply antimagic to magic characters (an escape character)	*/
		for(a= magic; *a; ++a) if(*l == *a) break;
		if(*a) *(b++)= '\\';
		}

	if(l[0] == '/') break;					/* no slashes (comments) allowed	*/
	*b= *l;									/* copy character into buffer		*/
	}
if(!*l) *(b++)= '$';						/* terminate buffer with '$'		*/
*b= '\0';									/* terminate buffer with null		*/

return buf;
}

/* --------------------------------------------------------------------- */
/* mergetags: read in <tags> file, generate more Labels {{{2 */
static long mergetags(void)
{
char  buf[BUFSIZE];
char  label[BUFSIZE];
char  filename[BUFSIZE];
char *pattern           = NULL;
long  tagcnt            = 0;
FILE *fp                = NULL;


/* read in <tags> and use makeLabel */
if(mflag) {
#ifdef vms
	fp= fopen(tagfile,"r");
#else
	fp= fopen(tagfile,"r");
#endif
	if(fp) {
		while(fgets(buf,BUFSIZE,fp)) {
			++tagcnt;
			buf[strlen(buf)-1]= '\0';
			sscanf(buf,"%s %s",label,filename);
			pattern= stpnxt(buf,"%s%s%b");
			makeLabel(label,pattern,(int) ZeroState,TTctags);
			stralloc(ltl->fname,filename,"mergetags");
			}
		fclose(fp);
		}
	}

return tagcnt;
}

/* --------------------------------------------------------------------- */
/* cppaccept: this function is used by parsrch/hdrbase to determine what {{{2
 * characters can be accepted as part of a potential C++ variable
 */
int cppaccept(int c)
{
return c == '_' || isalnum(c) || c == ':' || c == '~';
}

/* --------------------------------------------------------------------- */
/* texaccept: this function is used by parsrch/texbase to determine what {{{2
 * characters can be accepted as a potential TeX/LaTeX macro
 */
int texaccept(int c)
{
return isalpha(c);
}

/* --------------------------------------------------------------------- */
/* ShowFilemode: this function {{{2 */

/* --------------------------------------------------------------------- */
/* vim: ts=4 fdm=marker
 */
